from __future__ import annotations

import datetime as dt
import json
from functools import lru_cache

from .. import shared_ctx
from .._logging import logger
from .._progress_reporting import make_default_progress
from .._utils.perf import time_op
from . import cataloguer

_LOAD_CATALOGUE_LOCK = '_LOAD_CATALOGUE_'

_base_catalogue_url = (
    f'https://raw.githubusercontent.com/layday/instawow-data/data/'
    f'base-catalogue-v{cataloguer.CATALOGUE_VERSION}.compact.json'
)
_catalogue_ttl = dt.timedelta(hours=4)


@lru_cache(1)
def _parse_catalogue(raw_catalogue: bytes):
    with time_op(lambda t: logger.debug(f'parsed catalogue in {t:.3f}s')):
        return cataloguer.ComputedCatalogue.from_base_catalogue(
            json.loads(raw_catalogue),
        )


async def synchronise() -> cataloguer.ComputedCatalogue:
    "Fetch the catalogue from the interwebs and load it."
    async with (
        shared_ctx.locks[_LOAD_CATALOGUE_LOCK],
        shared_ctx.web_client.get(
            _base_catalogue_url,
            expire_after=_catalogue_ttl,
            raise_for_status=True,
            trace_request_ctx={
                'progress': make_default_progress(
                    type_='download', label='Synchronising catalogue'
                )
            },
        ) as response,
    ):
        raw_catalogue = await response.read()

    return _parse_catalogue(raw_catalogue)



from __future__ import annotations

import posixpath
import zipfile
from collections.abc import Callable, Iterable, Set
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from typing import NamedTuple, Protocol


class ArchiveOpener(Protocol):  # pragma: no cover
    def __call__(self, archive_path: Path) -> AbstractContextManager[Archive]: ...


class Archive(NamedTuple):
    top_level_folders: Set[str]
    extract: Callable[[Path], None]


def find_archive_addon_tocs(names: Iterable[str]):
    "Find top-level folders in a list of archive member paths."
    for name in names:
        if name.count(posixpath.sep) == 1:
            head, tail = posixpath.split(name)
            if tail.startswith(head) and tail[-4:].lower() == '.toc':
                yield (name, head)


def make_archive_member_filter_fn(base_dirs: Set[str]):
    "Filter out items which are not sub-paths of top-level folders in an archive."

    def is_subpath(name: str):
        head, sep, _ = name.partition(posixpath.sep)
        return head in base_dirs if sep else False

    return is_subpath


@contextmanager
def open_zip_archive(archive_path: Path):
    with zipfile.ZipFile(archive_path) as archive:
        names = archive.namelist()
        top_level_folders = {h for _, h in find_archive_addon_tocs(names)}

        def extract(parent_path: Path) -> None:
            should_extract = make_archive_member_filter_fn(top_level_folders)
            archive.extractall(parent_path, members=(n for n in names if should_extract(n)))

        yield Archive(top_level_folders, extract)



from __future__ import annotations

from pathlib import PurePath


def is_file_uri(uri: str) -> bool:
    return uri.startswith('file://')


def file_uri_to_path(file_uri: str) -> str:
    "Convert a file URI to a path that works both on Windows and *nix."
    from urllib.parse import unquote

    unprefixed_path = unquote(file_uri.removeprefix('file://'))
    # A slash is prepended to the path even when there isn't one there
    # on Windows.  The ``Path`` instance will inherit from either
    # ``PurePosixPath`` or ``PureWindowsPath``; this will be a no-op on POSIX.
    if PurePath(unprefixed_path[1:]).drive:
        unprefixed_path = unprefixed_path[1:]
    return unprefixed_path


def as_plain_text_data_url(body: str = '') -> str:
    from urllib.parse import quote

    return f'data:,{quote(body)}'


def extract_byte_range_offset(content_range: str) -> int:
    return int(content_range.replace('bytes ', '').partition('-')[0])


def open_url(url: str) -> None:
    import click

    click.launch(url)



from __future__ import annotations

import contextlib
import os
import sqlite3
import threading
from collections.abc import Iterator
from typing import TypeAlias

from .._utils.iteration import WeakValueDefaultDictionary

Connection: TypeAlias = sqlite3.Connection
Row: TypeAlias = sqlite3.Row


_connection_lock_factory = WeakValueDefaultDictionary[object, threading.Lock](threading.Lock)


_VERSION = 1

_SCHEMA = f"""
CREATE TABLE pkg (
    source VARCHAR NOT NULL,
    id VARCHAR NOT NULL,
    slug VARCHAR NOT NULL,
    name VARCHAR NOT NULL,
    description VARCHAR NOT NULL,
    url VARCHAR NOT NULL,
    download_url VARCHAR NOT NULL,
    date_published DATETIME NOT NULL,
    version VARCHAR NOT NULL,
    changelog_url VARCHAR NOT NULL,
    PRIMARY KEY (source, id)
);

CREATE TABLE pkg_version_log (
    version VARCHAR NOT NULL,
    install_time DATETIME DEFAULT (CURRENT_TIMESTAMP) NOT NULL,
    pkg_source VARCHAR NOT NULL,
    pkg_id VARCHAR NOT NULL,
    PRIMARY KEY (version, pkg_source, pkg_id)
);
CREATE INDEX pkg_version_log_faux_fk ON pkg_version_log (pkg_source, pkg_id);

CREATE TABLE pkg_options (
    any_flavour BOOLEAN NOT NULL,
    any_release_type BOOLEAN NOT NULL,
    version_eq BOOLEAN NOT NULL,
    pkg_source VARCHAR NOT NULL,
    pkg_id VARCHAR NOT NULL,
    PRIMARY KEY (pkg_source, pkg_id),
    CONSTRAINT fk_pkg_options_pkg_source_and_id
        FOREIGN KEY (pkg_source, pkg_id)
        REFERENCES pkg (source, id)
        ON DELETE CASCADE
);
CREATE UNIQUE INDEX pkg_options_fk ON pkg_options (pkg_source, pkg_id);

CREATE TABLE pkg_folder (
    name VARCHAR NOT NULL,
    pkg_source VARCHAR NOT NULL,
    pkg_id VARCHAR NOT NULL,
    PRIMARY KEY (name),
    CONSTRAINT fk_pkg_folder_pkg_source_and_id
        FOREIGN KEY (pkg_source, pkg_id)
        REFERENCES pkg (source, id)
        ON DELETE CASCADE
);
CREATE INDEX pkg_folder_fk ON pkg_folder (pkg_source, pkg_id);

CREATE TABLE pkg_dep (
    id VARCHAR NOT NULL,
    pkg_source VARCHAR NOT NULL,
    pkg_id VARCHAR NOT NULL,
    PRIMARY KEY (id, pkg_source, pkg_id),
    CONSTRAINT fk_pkg_dep_pkg_source_and_id
        FOREIGN KEY (pkg_source, pkg_id)
        REFERENCES pkg (source, id)
        ON DELETE CASCADE
);
CREATE INDEX pkg_dep_fk ON pkg_dep (pkg_source, pkg_id);

PRAGMA user_version = {_VERSION};
"""


def _get_version(connection: Connection) -> int | None:
    has_pkg_table = connection.execute('PRAGMA table_info("pkg")').fetchone()
    if has_pkg_table:
        (user_version,) = connection.execute('PRAGMA user_version').fetchone()
        return user_version


def _create(connection: Connection):
    with connection as transaction:
        transaction.executescript(_SCHEMA)


def _migrate(connection: Connection, current_version: int, new_version: int):
    from ._migrations import MIGRATIONS

    # SQLite migration reference:
    # https://www.sqlite.org/lang_altertable.html#otheralter

    with (
        connection as transaction,  # Steps 2 and 11.
        contextlib.ExitStack() as exit_stack,
    ):
        # Steps 1 and 12.
        transaction.execute('PRAGMA foreign_keys = OFF')
        exit_stack.callback(lambda: transaction.execute('PRAGMA foreign_keys = ON'))

        if new_version > current_version:
            for intermediate_version in range(current_version + 1, new_version + 1):
                MIGRATIONS[intermediate_version]().upgrade(transaction)

        else:
            for intermediate_version in range(current_version, new_version, -1):
                MIGRATIONS[intermediate_version]().downgrade(transaction)

        # Step 10.
        transaction.execute('PRAGMA foreign_key_check')

        # Stamp database with new version.
        transaction.execute(f'PRAGMA user_version = {new_version}')


def _configure(connection: Connection):
    connection.execute('PRAGMA foreign_keys = ON')
    # connection.set_trace_callback(print)


def _prepare_database(path: os.PathLike[str]) -> sqlite3.Connection:
    connection = sqlite3.connect(path, check_same_thread=False)
    connection.row_factory = sqlite3.Row
    _configure(connection)

    current_version = _get_version(connection)
    if current_version is not None and current_version != _VERSION:
        _migrate(connection, current_version, _VERSION)
    elif current_version is None:
        _create(connection)

    return connection


class DatabaseHandle:
    def __init__(self, path: os.PathLike[str]) -> None:
        self.path = path
        self._connection = None

    @contextlib.contextmanager
    def connect(self) -> Iterator[sqlite3.Connection]:
        if not self._connection:
            with _connection_lock_factory[self.path]:
                if not self._connection:
                    self._connection = _prepare_database(self.path)

        yield self._connection

    def close(self) -> None:
        if self._connection:
            with _connection_lock_factory[self.path]:
                if self._connection:
                    self._connection.close()
                    self._connection = None

    @classmethod
    @contextlib.contextmanager
    def transact(cls, connection: Connection) -> Iterator[sqlite3.Connection]:
        with connection:
            yield connection

    @classmethod
    @contextlib.contextmanager
    def use_tuple_factory(cls, connection: Connection) -> Iterator[sqlite3.Cursor]:
        with contextlib.closing(connection.cursor()) as cursor:
            cursor.row_factory = None
            yield cursor



from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import datetime, timezone

from .. import results as R
from .._utils.aio import run_in_thread
from ..catalogue.cataloguer import CatalogueEntry
from ..definitions import ChangelogFormat, Defn, SourceMetadata
from ..resolvers import BaseResolver, PkgCandidate
from ..wow_installations import Flavour

_ADDONS = {
    ('0', 'weakauras-companion'),
    ('1', 'weakauras-companion-autoupdate'),
}


class InstawowResolver(BaseResolver):
    metadata = SourceMetadata(
        id='instawow',
        name='instawow',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key=None,
    )
    requires_access_token = None

    async def _resolve_one(self, defn: Defn, metadata: None) -> PkgCandidate:
        try:
            source_id, slug = next(p for p in _ADDONS if defn.alias in p)
        except StopIteration:
            raise R.PkgNonexistent from None

        from instawow_wa_updater._core import WaCompanionBuilder

        builder = WaCompanionBuilder(self._config)
        if source_id == '1':
            await builder.build()

        return PkgCandidate(
            id=source_id,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=builder.config.addon_zip_file.as_uri(),
            date_published=datetime.now(timezone.utc),
            version=await run_in_thread(builder.get_version)(),
            changelog_url=builder.config.changelog_file.as_uri(),
        )

    @classmethod
    async def catalogue(cls) -> AsyncIterator[CatalogueEntry]:
        yield CatalogueEntry(
            source=cls.metadata.id,
            id='1',
            slug='weakauras-companion-autoupdate',
            name='WeakAuras Companion',
            url='https://github.com/layday/instawow',
            game_flavours=frozenset(Flavour),
            download_count=1,
            last_updated=datetime.now(timezone.utc),
            folders=[
                frozenset({'WeakAurasCompanion'}),
            ],
        )



from __future__ import annotations

from .pkg_db._models import Pkg as Pkg
from .pkg_db._models import PkgDep as PkgDep
from .pkg_db._models import PkgFolder as PkgFolder
from .pkg_db._models import PkgLoggedVersion as PkgLoggedVersion
from .pkg_db._models import PkgOptions as PkgOptions
from .pkg_db._models import make_db_converter as make_db_converter
from .pkg_db._ops import build_pkg_from_row_mapping as build_pkg_from_row_mapping



from __future__ import annotations

import json
import shutil
from collections.abc import Callable
from functools import partial
from textwrap import dedent
from typing import TypeAlias
from unittest import mock

import pytest
from cattrs.preconf.json import make_converter as make_json_converter
from click.testing import CliRunner, Result
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import PipeInput, create_pipe_input
from prompt_toolkit.output import DummyOutput

from instawow import __version__
from instawow.cli import cli
from instawow.config import ProfileConfig
from instawow.definitions import SourceMetadata

Run: TypeAlias = Callable[[str], Result]


@pytest.fixture(autouse=True, scope='module')
def _mock_pt_progress_bar():
    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr('prompt_toolkit.shortcuts.progress_bar.ProgressBar', mock.MagicMock())
    yield
    monkeypatch.undo()


@pytest.fixture
def pt_input():
    with (
        create_pipe_input() as pipe_input,
        create_app_session(input=pipe_input, output=DummyOutput()),
    ):
        yield pipe_input


@pytest.fixture
async def run(
    monkeypatch: pytest.MonkeyPatch,
    iw_profile_config: ...,
):
    import asyncio

    loop = asyncio.get_running_loop()
    monkeypatch.setattr('asyncio.run', loop.run_until_complete)

    return partial(CliRunner().invoke, cli, catch_exceptions=False)


@pytest.fixture
def install_molinari_and_run(
    run: Run,
):
    run('install curse:molinari')
    return run


@pytest.fixture
def pretend_install_molinari_and_run(
    iw_profile_config: ProfileConfig,
    run: Run,
):
    molinari = iw_profile_config.addon_dir / 'Molinari'
    molinari.mkdir()
    (molinari / 'Molinari.toc').write_text(
        """\
## X-Curse-Project-ID: 20338
## X-WoWI-ID: 13188
"""
    )
    return run


@pytest.mark.parametrize(
    'alias',
    [
        'curse:molinari',
        'wowi:13188-molinari',
        'tukui:tukui',
        'github:p3lim-wow/Molinari',
    ],
)
def test_valid_pkg_lifecycle(
    run: Run,
    alias: str,
):
    assert run(f'install {alias}').output.startswith(f'✓ {alias}\n  installed')
    assert run(f'install {alias}').output == f'✗ {alias}\n  package already installed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is up to date\n'
    assert run(f'remove {alias}').output == f'✓ {alias}\n  removed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['instawow:gargantuan-wigs'])
def test_nonexistent_pkg_lifecycle(
    run: Run,
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ {alias}\n  package does not exist\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['foo:bar'])
def test_invalid_source_lifecycle(
    run: Run,
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ :{alias}\n  package source is invalid\n'
    assert run(f'update {alias}').output == f'✗ :{alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ :{alias}\n  package is not installed\n'


def test_reconciled_folder_conflict_on_install(
    run: Run,
):
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert run('install wowi:13188-molinari').output == (
        '✗ wowi:13188-molinari\n'
        '  package folders conflict with installed package Molinari\n'
        '    (curse:20338)\n'
    )


def test_unreconciled_folder_conflict_on_install(
    iw_profile_config: ProfileConfig,
    run: Run,
):
    iw_profile_config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == "✗ curse:molinari\n  package folders conflict with 'Molinari'\n"
    )
    assert run('install --replace curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed'
    )


def test_keep_folders_on_remove(
    iw_profile_config: ProfileConfig,
    install_molinari_and_run: Run,
):
    assert (
        install_molinari_and_run('remove --keep-folders curse:molinari').output
        == '✓ curse:molinari\n  removed\n'
    )
    assert iw_profile_config.addon_dir.joinpath('Molinari').is_dir()


def test_version_strategy_lifecycle(
    run: Run,
):
    assert run('install curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed 100205.111-Release'
    )
    assert (
        run('install curse:molinari#version_eq=foo').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is up to date\n'
    assert run('update curse:molinari#version_eq=100005.97-Release').output == dedent(
        """\
        ✓ curse:molinari
          updated 100205.111-Release to 100005.97-Release with new strategies:
            version_eq='100005.97-Release'
        """
    )
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert run('install curse:molinari#version_eq=foo').output == dedent(
        """\
        ✗ curse:molinari
          no files found for: any_flavour=None; any_release_type=None;
            version_eq='foo'
        """
    )
    assert (
        run('install curse:molinari#version_eq=100005.97-Release').output
        == '✓ curse:molinari\n  installed 100005.97-Release\n'
    )
    assert run('update').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('update curse:molinari#=').output == dedent(
        """\
        ✓ curse:molinari
          updated 100005.97-Release to 100205.111-Release with new strategies:
            version_eq=None
        """
    )
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


def test_install_options(
    run: Run,
):
    assert run('install curse:molinari#any_release_type,any_flavour').output == dedent(
        """\
        ✓ curse:molinari
          installed 100205.111-Release
        """
    )


@pytest.mark.parametrize('step', [1, -1])
def test_install_order_is_respected(
    run: Run,
    step: int,
):
    assert run(
        'install '
        + ' '.join(
            [
                'curse:molinari',
                'curse:molinari#version_eq=100005.97-Release',
            ][::step]
        )
    ).output == dedent(
        f"""\
        ✓ curse:molinari
          installed {'100205.111-Release' if step == 1 else '100005.97-Release'}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        """
    )


def test_install_invalid_defn(
    run: Run,
):
    result = run('install foo')
    assert result.exit_code == 2
    assert result.output.endswith("Error: Invalid value for '[ADDONS]...': foo\n")


def test_install_dry_run(
    run: Run,
):
    assert run('install --dry-run curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          would have installed 100205.111-Release
        """
    )


def test_debug_config(
    iw_profile_config: ProfileConfig,
    run: Run,
):
    assert (
        run('debug config').output
        == json.dumps(iw_profile_config.unstructure_for_display(), indent=2) + '\n'
    )


def test_debug_sources(
    run: Run,
):
    json_converter = make_json_converter()

    output = run('debug sources').output
    source_metadata = json_converter.loads(output, list[SourceMetadata])
    assert len(source_metadata) > 1


@pytest.mark.parametrize('command', ['configure', 'list'], ids=['explicit', 'implicit'])
def test_configure__create_new_profile(
    monkeypatch: pytest.MonkeyPatch,
    pt_input: PipeInput,
    iw_profile_config: ProfileConfig,
    run: Run,
    command: str,
):
    # In case there is a WoW installation in the test environment.
    monkeypatch.setattr('instawow.wow_installations.find_installations', lambda: iter([]))

    pt_input.send_text(f'{iw_profile_config.addon_dir}\r\rY\r')
    assert run(f'-p foo {command}').output == (
        'Navigate to https://github.com/login/device and paste the code below:\n'
        '  WDJB-MJHT\n'
        'Waiting...\n'
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/foo/config.json"}\n'
    )


def test_configure__update_existing_profile_interactively(
    pt_input: PipeInput,
    iw_profile_config: ProfileConfig,
    run: Run,
):
    pt_input.send_text('Y\r')
    assert run('configure global_config.auto_update_check').output == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/__default__/config.json"}\n'
    )


def test_configure__update_existing_profile_directly(
    iw_profile_config: ProfileConfig,
    run: Run,
):
    assert run('configure global_config.auto_update_check=0').output == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_profile_config.global_config.config_dir / "profiles/__default__/config.json"}\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__pkg_not_installed(
    run: Run,
    options: str,
):
    assert (
        run(f'rollback {options} curse:molinari').output
        == '✗ curse:molinari\n  package is not installed\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__unsupported(
    run: Run,
    options: str,
):
    assert run('install wowi:13188-molinari').exit_code == 0
    assert (
        run(f'rollback {options} wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  strategies are not valid for source: version_eq\n'
    )


def test_rollback__single_version(
    run: Run,
):
    assert run('install curse:molinari').exit_code == 0
    assert (
        run('rollback curse:molinari').output == '✗ curse:molinari\n  cannot find older versions\n'
    )


def test_rollback__multiple_versions(
    pt_input: PipeInput,
    run: Run,
):
    assert run('install curse:molinari#version_eq=100005.97-Release').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    pt_input.send_text('\r\r')
    assert run('rollback curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          updated 100205.111-Release to 100005.97-Release with new strategies:
            version_eq='100005.97-Release'
        """
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__rollback_multiple_versions(
    pt_input: PipeInput,
    run: Run,
    options: str,
):
    assert run('install curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari#version_eq=100005.97-Release').exit_code == 0
    pt_input.send_text('\r\r')
    assert run(f'rollback {options} curse:molinari').output == dedent(
        """\
        ✓ curse:molinari
          updated 100005.97-Release to 100205.111-Release with new strategies:
            version_eq=None
        """
        if options == '--undo'
        else """\
        ✓ curse:molinari
          updated 100005.97-Release to 100205.111-Release
        """
    )


def test_reconcile__list_unreconciled(
    pretend_install_molinari_and_run: Run,
):
    assert (
        pretend_install_molinari_and_run('reconcile --list-unreconciled').output
        == (
            'unreconciled\n'  # fmt: skip
            '------------\n'
            'Molinari    \n'
        )
    )


def test_reconcile_leftovers(
    pt_input: PipeInput,
    pretend_install_molinari_and_run: Run,
):
    pt_input.send_text('sss')  # Skip
    assert pretend_install_molinari_and_run(
        'reconcile'
    ).output.endswith(
        'unreconciled\n'  # fmt: skip
        '------------\n'
        'Molinari    \n'
    )


def test_reconcile__auto_reconcile(
    pretend_install_molinari_and_run: Run,
):
    assert pretend_install_molinari_and_run('reconcile --auto').output == dedent(
        """\
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_reconcile__abort_interactive_reconciliation(
    pt_input: PipeInput,
    pretend_install_molinari_and_run: Run,
):
    pt_input.send_text('\x03')  # ^C
    assert pretend_install_molinari_and_run('reconcile').output.endswith('Aborted!\n')


def test_reconcile__complete_interactive_reconciliation(
    pt_input: PipeInput,
    pretend_install_molinari_and_run: Run,
):
    pt_input.send_text('\ry')
    assert pretend_install_molinari_and_run('reconcile').output.endswith(
        dedent(
            """\
            ✓ github:p3lim-wow/molinari
              installed 100205.111-Release
            """
        )
    )


def test_reconcile__reconciliation_complete(
    run: Run,
):
    assert run('reconcile').output == 'No add-ons left to reconcile.\n'


def test_rereconcile(
    pt_input: PipeInput,
    install_molinari_and_run: Run,
):
    pt_input.send_text('\ry')
    assert install_molinari_and_run('rereconcile').output == dedent(
        """\
        ✓ curse:molinari
          removed
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_rereconcile_with_args(
    pt_input: PipeInput,
    install_molinari_and_run: Run,
):
    pt_input.send_text('\ry')
    assert install_molinari_and_run('rereconcile foo').output == ''

    pt_input.send_text('\ry')
    assert install_molinari_and_run('rereconcile molinari').output == dedent(
        """\
        ✓ curse:molinari
          removed
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        """
    )


def test_search__no_results(
    run: Run,
):
    assert run('search ∅').output == 'No results found.\n'


def test_search__exit_without_selecting(
    pt_input: PipeInput,
    run: Run,
):
    pt_input.send_text('\r')  # enter
    assert run('search molinari').output == (
        'Nothing was selected; select add-ons with <space> and confirm by pressing <enter>.\n'
    )


def test_search__exit_after_selection(
    pt_input: PipeInput,
    run: Run,
):
    pt_input.send_text(' \rn')  # space, enter, "n"
    assert run('search molinari').output == ''


def test_search__install_one(
    pt_input: PipeInput,
    run: Run,
):
    pt_input.send_text(' \ry')  # space, enter, enter
    assert run('search molinari --source curse').output == dedent(
        """\
        ✓ curse:molinari
          installed 100205.111-Release
        """
    )


def test_search__install_multiple_conflicting(
    pt_input: PipeInput,
    run: Run,
):
    pt_input.send_text(' \x1b[B \ry')  # space, arrow down, space, enter, enter
    assert run('search molinari').output == dedent(
        """\
        ✓ github:p3lim-wow/molinari
          installed 100205.111-Release
        ✗ wowi:13188
          package folders conflict with installed package Molinari
            (github:388670)
        """
    )


def test_changelog_output_installed_convert(
    install_molinari_and_run: Run,
):
    output = (
        'curse:molinari:\n  Changes in 90200.82-Release:'
        if shutil.which('pandoc')
        else 'curse:molinari:\n  <h3>Changes in'
    )
    assert install_molinari_and_run('view-changelog curse:molinari').output.startswith(output)


def test_changelog_output_installed_no_convert(
    install_molinari_and_run: Run,
):
    assert install_molinari_and_run(
        'view-changelog --no-convert curse:molinari'
    ).output.startswith('curse:molinari:\n  <h3>Changes in')


def test_changelog_output_not_installed(
    run: Run,
):
    assert run('view-changelog curse:molinari').output == ''


def test_changelog_output_argless(
    install_molinari_and_run: Run,
):
    assert install_molinari_and_run('view-changelog').output.startswith('curse:molinari:')


def test_changelog_output_remote(
    run: Run,
):
    assert run('view-changelog --remote curse:molinari').output.startswith('curse:molinari:')


@pytest.mark.parametrize(
    ('command', 'exit_code'),
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(
    monkeypatch: pytest.MonkeyPatch,
    install_molinari_and_run: Run,
    command: str,
    exit_code: int,
):
    monkeypatch.setattr('instawow._utils.file.reveal_folder', lambda *_, **__: ...)
    assert install_molinari_and_run(command).exit_code == exit_code


def test_can_list_with_substr_match(
    install_molinari_and_run: Run,
):
    assert install_molinari_and_run('list mol').output == 'curse:molinari\n'
    assert install_molinari_and_run('list foo').output == ''
    assert install_molinari_and_run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(install_molinari_and_run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export(
    install_molinari_and_run: Run,
):
    output = install_molinari_and_run('list -f json').output
    assert json.loads(output)[0]['name'] == 'Molinari'


def test_show_version(
    run: Run,
):
    assert run('--version').output == f'instawow, version {__version__}\n'


def test_plugin_hook_command_can_be_invoked(
    run: Run,
):
    pytest.importorskip('instawow_test_plugin')
    assert run('plugins foo').output == 'success!\n'



from __future__ import annotations

import asyncio
import contextvars
from collections.abc import Awaitable, Callable, Mapping
from contextlib import contextmanager
from itertools import count
from typing import Any, Generic, Literal, TypeAlias, cast

from typing_extensions import Never, NotRequired, TypedDict, TypeVar

_T = TypeVar('_T')

_TProgressType = TypeVar('_TProgressType', bound=str)
_TProgressUnit = TypeVar('_TProgressUnit', bound=str | None, default=None)


class Progress(TypedDict, Generic[_TProgressType, _TProgressUnit]):
    type_: _TProgressType
    unit: NotRequired[_TProgressUnit]
    label: NotRequired[str]
    current: int
    total: int


class GenericProgress(Progress[Literal['generic']]):
    pass


class DownloadProgress(Progress[Literal['download'], Literal['bytes']]):
    pass


def make_default_progress(*, type_: Literal['generic', 'download'], label: str, total: int = 0):
    match type_:
        case 'generic':
            return GenericProgress(
                type_='generic',
                label=label,
                current=0,
                total=total,
            )
        case 'download':
            return DownloadProgress(
                type_='download',
                unit='bytes',
                label=label,
                current=0,
                total=total,
            )


_TProgress = TypeVar('_TProgress', bound=Progress[Any, Any], default=Never)
ReadOnlyProgressGroup: TypeAlias = Mapping[int, DownloadProgress | GenericProgress | _TProgress]

progress_notifiers = contextvars.ContextVar(
    'progress_notifiers',
    default=set[Callable[[int, Progress[Any, Any] | Literal['unset']], None]](),
)

_progress_id_gen = count()


def get_next_progress_id() -> int:
    return next(_progress_id_gen)


class make_progress_receiver(Generic[_TProgress]):
    @contextmanager
    def __new__(cls):
        asyncio.get_running_loop()  # Raise if constructed outside async context.

        emit_event = asyncio.Event()

        progress_group: dict[int, Progress[Any, Any]] = {}

        def waken(progress_id: int, progress: Progress[Any, Any] | Literal['unset']):
            if progress == 'unset':
                try:
                    del progress_group[progress_id]
                except KeyError:
                    pass
            else:
                progress_group[progress_id] = progress

            emit_event.set()

        progress_notifiers.set(progress_notifiers.get() | {waken})

        try:

            async def receive():
                while True:
                    await emit_event.wait()
                    yield cast(ReadOnlyProgressGroup[_TProgress], progress_group)
                    emit_event.clear()

            yield receive()

        finally:
            progress_notifiers.set(progress_notifiers.get() - {waken})


def update_progress(progress_id: int, progress: Progress[Any, Any] | Literal['unset']) -> None:
    for notify in progress_notifiers.get():
        notify(progress_id, progress)


def make_incrementing_progress_tracker(total: int, label: str):
    progress_id = get_next_progress_id()
    progress = make_default_progress(type_='generic', total=total, label=label)

    def do_update_progress():
        update_progress(
            progress_id,
            (
                'unset'
                if progress['total'] == 0 or progress['current'] == progress['total']
                else progress
            ),
        )

    def on_done(_task: asyncio.Task[object]):
        progress['current'] += 1
        do_update_progress()

    def track(awaitable: Awaitable[_T]) -> Awaitable[_T]:
        future = asyncio.ensure_future(awaitable)
        future.add_done_callback(on_done)
        return future

    do_update_progress()
    return track



from __future__ import annotations

import contextlib
from collections.abc import Mapping
from typing import Protocol

from . import Connection


class Migration(Protocol):
    def upgrade(self, connection: Connection) -> None: ...

    def downgrade(self, connection: Connection) -> None: ...


class _BaseMigration(Migration, Protocol):
    @contextlib.contextmanager
    def _with_table_op(self, connection: Connection):
        yield


class _Migration_1(_BaseMigration):
    def upgrade(self, connection: Connection) -> None:
        connection.execute(
            'CREATE UNIQUE INDEX pkg_options_fk ON pkg_options (pkg_source, pkg_id)',
        )
        connection.execute(
            'CREATE INDEX pkg_folder_fk ON pkg_folder (pkg_source, pkg_id)',
        )
        connection.execute(
            'CREATE INDEX pkg_dep_fk ON pkg_dep (pkg_source, pkg_id)',
        )
        connection.execute(
            'CREATE INDEX pkg_version_log_faux_fk ON pkg_version_log (pkg_source, pkg_id)',
        )

    def downgrade(self, connection: Connection) -> None:
        connection.execute(
            'DROP INDEX pkg_options_fk',
        )
        connection.execute(
            'DROP INDEX pkg_folder_fk',
        )
        connection.execute(
            'DROP INDEX pkg_dep_fk',
        )
        connection.execute(
            'DROP INDEX pkg_version_log_faux_fk',
        )


MIGRATIONS: Mapping[int, type[Migration]] = dict(
    enumerate(
        [
            _Migration_1,
        ],
        start=1,
    )
)



from __future__ import annotations

import click

import instawow.plugins


@click.command('gui')
@click.pass_context
def _gui_command(ctx: click.Context) -> None:
    "Fire up the GUI."
    from instawow import __version__
    from instawow._logging import setup_logging
    from instawow.config import GlobalConfig, ProfileConfig

    from .app import make_app

    global_config = GlobalConfig.read().ensure_dirs()
    dummy_jsonrpc_config = ProfileConfig.make_dummy_config(
        global_config=global_config, profile='__jsonrpc__'
    ).ensure_dirs()

    params = ctx.find_root().params
    setup_logging(dummy_jsonrpc_config.logging_dir, *params['verbose'])

    make_app(version=__version__).main_loop()


@instawow.plugins.hookimpl
def instawow_add_commands():
    return (_gui_command,)



from __future__ import annotations

import asyncio
import concurrent.futures
import contextvars
import enum
import importlib.resources
import json
import os
from collections.abc import Callable, Iterator, Set
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from datetime import datetime
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, Literal, TypeVar

import aiohttp
import aiohttp.typedefs
import aiohttp.web
import attrs
import cattrs
import cattrs.preconf.json
import toga
from aiohttp_rpc import JsonRpcMethod
from aiohttp_rpc import middlewares as rpc_middlewares
from aiohttp_rpc.errors import InvalidParams, ServerError
from aiohttp_rpc.server import WsJsonRpcServer
from typing_extensions import ParamSpec, TypedDict
from yarl import URL

from instawow import __version__, matchers, pkg_management, pkg_models, shared_ctx
from instawow import results as R
from instawow._logging import logger
from instawow._progress_reporting import ReadOnlyProgressGroup, make_progress_receiver
from instawow._utils.aio import cancel_tasks, gather, run_in_thread
from instawow._utils.datetime import datetime_fromisoformat
from instawow._utils.file import reveal_folder
from instawow._utils.iteration import WeakValueDefaultDictionary, uniq
from instawow._utils.web import open_url
from instawow._version_check import is_outdated
from instawow.catalogue.cataloguer import ComputedCatalogueEntry
from instawow.catalogue.search import search
from instawow.config import GlobalConfig, ProfileConfig, SecretStr, config_converter
from instawow.definitions import Defn, SourceMetadata, Strategies
from instawow.github_auth import get_codes, poll_for_access_token
from instawow.http import init_web_client
from instawow.wow_installations import Flavour, infer_flavour_from_addon_dir

from . import frontend

_T = TypeVar('_T')
_P = ParamSpec('_P')

_toga_handle = contextvars.ContextVar[toga.App]('_toga_handle')

_LOCK_PREFIX = object()


class _LockOperation(tuple[object, str], enum.Enum):
    ModifyProfile = (_LOCK_PREFIX, '_MODIFY_PROFILE_')
    UpdateGlobalConfig = (_LOCK_PREFIX, '_UPDATE_GLOBAL_CONFIG')
    InitiateGithubAuthFlow = (_LOCK_PREFIX, '_INITIATE_GITHUB_AUTH_FLOW_')


LOCALHOST = '127.0.0.1'

_converter = cattrs.Converter(
    unstruct_collection_overrides={
        Set: sorted,
    }
)
cattrs.preconf.json.configure_converter(_converter)
_converter.register_structure_hook(Path, lambda v, _: Path(v))
_converter.register_structure_hook(datetime, lambda v, _: datetime_fromisoformat(v))
_converter.register_structure_hook(Strategies, lambda v, _: Strategies(v))
_converter.register_unstructure_hook(Path, str)
_converter.register_unstructure_hook(Strategies, Strategies)


class _ConfigError(ServerError):
    code = -32001
    message = 'invalid configuration parameters'


class _ValidationErrorResponse(TypedDict):
    path: tuple[str | int, ...]
    message: str


def _transform_validation_errors(
    exc: cattrs.ClassValidationError | cattrs.IterableValidationError | BaseException,
    path: tuple[str | int, ...] = (),
) -> Iterator[_ValidationErrorResponse]:
    match exc:
        case cattrs.IterableValidationError():
            with_notes, _ = exc.group_exceptions()
            for exc, note in with_notes:
                assert isinstance(note.index, str | int)  # Dummy assert for type checking.
                new_path = (*path, note.index)
                if isinstance(exc, cattrs.ClassValidationError | cattrs.IterableValidationError):
                    yield from _transform_validation_errors(exc, new_path)
                else:
                    yield {
                        'path': new_path,
                        'message': str(exc),
                    }
        case cattrs.ClassValidationError():
            with_notes, _ = exc.group_exceptions()
            for exc, note in with_notes:
                new_path = (*path, note.name)
                if isinstance(exc, cattrs.ClassValidationError | cattrs.IterableValidationError):
                    yield from _transform_validation_errors(exc, new_path)
                else:
                    yield {
                        'path': new_path,
                        'message': str(exc),
                    }
        case _:
            pass


@contextmanager
def _reraise_validation_errors(
    error_class: type[ServerError | InvalidParams] = ServerError,
    values: dict[Any, Any] | None = None,
) -> Iterator[None]:
    try:
        yield
    except BaseException as exc:
        logger.info(f'invalid request: {(values, exc)}')
        raise error_class(data=list(_transform_validation_errors(exc))) from exc


@run_in_thread
def _read_global_config() -> GlobalConfig:
    with _reraise_validation_errors(_ConfigError):
        return GlobalConfig.read().ensure_dirs()


@run_in_thread
def _read_profile_config(global_config: GlobalConfig, profile: str) -> ProfileConfig:
    with _reraise_validation_errors(_ConfigError):
        return ProfileConfig.read(global_config, profile).ensure_dirs()


_methods: list[tuple[str, type[BaseParams]]] = []


def _register_method(method: str):
    def inner(param_class: type[BaseParams]):
        _methods.append((method, param_class))
        return attrs.frozen(slots=False)(param_class)

    return inner


@attrs.frozen(slots=False)
class _ProfileParamMixin:
    profile: str


@attrs.frozen(slots=False)
class _DefnParamMixin:
    defns: list[Defn]


class BaseParams:
    @classmethod
    def bind(cls, method: str, config_ctxs: _ConfigBoundCtxCollection) -> JsonRpcMethod:
        async def respond(**kwargs: Any):
            with _reraise_validation_errors(InvalidParams, kwargs):
                self = _converter.structure(kwargs, cls)
            return await self.respond(config_ctxs)

        return JsonRpcMethod(respond, name=method)

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> Any:
        raise NotImplementedError


@_register_method('config/write_profile')
class WriteProfileConfigParams(_ProfileParamMixin, BaseParams):
    addon_dir: Path
    game_flavour: Flavour
    infer_game_flavour: bool

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> ProfileConfig:
        async with config_ctxs.locks[(*_LockOperation.ModifyProfile, self.profile)]:
            with _reraise_validation_errors(_ConfigError):
                config = config_converter.structure(
                    {
                        'global_config': await _read_global_config(),
                        'profile': self.profile,
                        'addon_dir': self.addon_dir,
                        'game_flavour': self.game_flavour,
                    },
                    ProfileConfig,
                )
                if self.infer_game_flavour:
                    config = attrs.evolve(
                        config,
                        game_flavour=infer_flavour_from_addon_dir(config.addon_dir)
                        or Flavour.Retail,
                    )
                await run_in_thread(config.write)()

            # Unload the corresponding ``Manager`` instance for the
            # config to be reloaded on the next request
            await config_ctxs.unload(config.profile)

            return config


@_register_method('config/read_profile')
class ReadProfileConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> ProfileConfig:
        return await _read_profile_config(config_ctxs.global_config, self.profile)


@_register_method('config/delete_profile')
class DeleteProfileConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> None:
        config_ctx = await config_ctxs.load(self.profile)

        async with config_ctxs.locks[(*_LockOperation.ModifyProfile, self.profile)]:
            await run_in_thread(config_ctx.config.delete)()
            await config_ctxs.unload(self.profile)


@_register_method('config/list_profiles')
class ListProfilesParams(BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> list[str]:
        return await run_in_thread(list[str])(config_ctxs.global_config.iter_profiles())


@_register_method('config/update_global')
class UpdateGlobalConfigParams(BaseParams):
    access_tokens: dict[str, str | None]

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> GlobalConfig:
        def update_global_config_cb(global_config: GlobalConfig):
            return attrs.evolve(
                global_config,
                access_tokens=attrs.evolve(
                    global_config.access_tokens,
                    **{k: t if t is None else SecretStr(t) for k, t in self.access_tokens.items()},
                ),
            )

        return await config_ctxs.update_global_config(update_global_config_cb)


@_register_method('config/read_global')
class ReadGlobalConfigParams(BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> GlobalConfig:
        return await _read_global_config()


class GithubCodesResponse(TypedDict):
    user_code: str
    verification_uri: str


@_register_method('config/initiate_github_auth_flow')
class InitiateGithubAuthFlowParams(BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> GithubCodesResponse:
        return await config_ctxs.initiate_github_auth_flow()


class GithubAuthFlowStatusResponse(TypedDict):
    status: Literal['success', 'failure']


@_register_method('config/query_github_auth_flow_status')
class QueryGithubAuthFlowStatusParams(BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> GithubAuthFlowStatusResponse:
        return {'status': await config_ctxs.wait_for_github_auth_completion()}


@_register_method('config/cancel_github_auth_flow')
class CancelGithubAuthFlowParams(BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> None:
        await config_ctxs.cancel_github_auth_polling()


@_register_method('sources/list')
class ListSourcesParams(_ProfileParamMixin, BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> dict[str, SourceMetadata]:
        config_ctx = await config_ctxs.load(self.profile)
        return {r.metadata.id: r.metadata for r in config_ctx.resolvers.values()}


@_register_method('list')
class ListInstalledParams(_ProfileParamMixin, BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> list[pkg_models.Pkg]:
        config_ctx = await config_ctxs.load(self.profile)

        with config_ctx.database.connect() as connection:
            return [
                pkg_models.build_pkg_from_row_mapping(connection, p)
                for p in connection.execute('SELECT * FROM pkg ORDER BY lower(name)')
            ]


@_register_method('search')
class SearchParams(_ProfileParamMixin, BaseParams):
    search_terms: str
    limit: int
    sources: set[str]
    start_date: datetime | None
    installed_only: bool

    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[ComputedCatalogueEntry]:
        config_ctx = await config_ctxs.load(self.profile)

        return await search(
            config_ctx,
            self.search_terms,
            limit=self.limit,
            sources=self.sources,
            start_date=self.start_date,
            filter_installed='include_only' if self.installed_only else 'ident',
        )


class SuccessResult(TypedDict):
    status: Literal['success']
    addon: pkg_models.Pkg


class ErrorResult(TypedDict):
    status: Literal['failure', 'error']
    message: str


@_register_method('resolve')
class ResolveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[SuccessResult | ErrorResult]:
        config_ctx = await config_ctxs.load(self.profile)

        def extract_source(defn: Defn):
            if not defn.source:
                match = pkg_management.get_alias_from_url(config_ctx, defn.alias)
                if match:
                    source, alias = match
                    defn = attrs.evolve(defn, source=source, alias=alias)
            return defn

        results = await pkg_management.resolve(config_ctx, list(map(extract_source, self.defns)))
        return [
            {'status': 'success', 'addon': r}
            if isinstance(r, pkg_models.Pkg)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('install')
class InstallParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    replace: bool

    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[SuccessResult | ErrorResult]:
        config_ctx = await config_ctxs.load(self.profile)

        results = await pkg_management.install(
            config_ctx, self.defns, replace_folders=self.replace
        )
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('update')
class UpdateParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[SuccessResult | ErrorResult]:
        config_ctx = await config_ctxs.load(self.profile)

        results = await pkg_management.update(config_ctx, self.defns)
        return [
            {'status': r.status, 'addon': r.new_pkg}
            if isinstance(r, R.PkgUpdated)
            else {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('remove')
class RemoveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    keep_folders: bool

    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[SuccessResult | ErrorResult]:
        config_ctx = await config_ctxs.load(self.profile)

        results = await pkg_management.remove(
            config_ctx, self.defns, keep_folders=self.keep_folders
        )
        return [
            {'status': r.status, 'addon': r.old_pkg}
            if isinstance(r, R.PkgRemoved)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('pin')
class PinParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[SuccessResult | ErrorResult]:
        config_ctx = await config_ctxs.load(self.profile)

        results = await pkg_management.pin(config_ctx, self.defns)
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('get_changelog')
class GetChangelogParams(_ProfileParamMixin, BaseParams):
    source: str
    changelog_url: str

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> str:
        config_ctx = await config_ctxs.load(self.profile)

        return await pkg_management.get_changelog(
            config_ctx, source=self.source, uri=self.changelog_url
        )


class AddonMatch(TypedDict):
    folders: list[AddonMatch_AddonFolder]
    matches: list[pkg_models.Pkg]


class AddonMatch_AddonFolder(TypedDict):
    name: str
    version: str


@_register_method('reconcile')
class ReconcileParams(_ProfileParamMixin, BaseParams):
    matcher: str

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> list[AddonMatch]:
        config_ctx = await config_ctxs.load(self.profile)

        leftovers = await run_in_thread(matchers.get_unreconciled_folders)(config_ctx)

        match_groups = await matchers.DEFAULT_MATCHERS[self.matcher](
            config_ctx, leftovers=leftovers
        )

        resolved_defns = await pkg_management.resolve(
            config_ctx, uniq(d for _, b in match_groups for d in b)
        )
        pkgs, _ = pkg_management.bucketise_results(resolved_defns.items())

        matched_folders = [
            (a, [i for i in (pkgs.get(d) for d in s) if i]) for a, s in match_groups
        ]
        unmatched_folders = (
            ([a], list[pkg_models.Pkg]())
            for a in sorted(leftovers - frozenset(i for a, _ in matched_folders for i in a))
        )
        return [
            {
                'folders': [{'name': f.name, 'version': f.toc_reader.version} for f in a],
                'matches': m,
            }
            for a, m in chain(matched_folders, unmatched_folders)
        ]


class ReconcileInstalledCandidate(TypedDict):
    installed_addon: pkg_models.Pkg
    alternative_addons: list[pkg_models.Pkg]


@_register_method('get_reconcile_installed_candidates')
class GetReconcileInstalledCandidatesParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[ReconcileInstalledCandidate]:
        config_ctx = await config_ctxs.load(self.profile)

        with config_ctx.database.connect() as connection:
            installed_pkgs = [
                pkg_models.build_pkg_from_row_mapping(connection, p)
                for p in connection.execute('SELECT * FROM pkg ORDER BY lower(name)')
            ]

        defn_groups = await pkg_management.find_equivalent_pkg_defns(config_ctx, installed_pkgs)
        resolved_defns = await pkg_management.resolve(
            config_ctx, uniq(d for b in defn_groups.values() for d in b)
        )
        pkgs, _ = pkg_management.bucketise_results(resolved_defns.items())
        return [
            {'installed_addon': p, 'alternative_addons': m}
            for p, s in defn_groups.items()
            for m in ([i for i in (pkgs.get(d) for d in s) if i],)
            if m
        ]


class DownloadProgressReport(TypedDict):
    defn: Defn
    progress: float


@_register_method('get_download_progress')
class GetDownloadProgressParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, config_ctxs: _ConfigBoundCtxCollection
    ) -> list[DownloadProgressReport]:
        config_ctx = await config_ctxs.load(self.profile)

        return [
            {'defn': p['defn'], 'progress': p['current'] / p['total']}
            for p in config_ctxs.current_progress.values()
            if p['type_'] == 'pkg_download'
            and p['total']
            and p['profile'] == config_ctx.config.profile
        ]


class GetVersionResult(TypedDict):
    installed_version: str
    new_version: str | None


@_register_method('meta/get_version')
class GetVersionParams(BaseParams):
    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> GetVersionResult:
        outdated, new_version = await is_outdated(config_ctxs.global_config)
        return {
            'installed_version': __version__,
            'new_version': new_version if outdated else None,
        }


@_register_method('assist/open_url')
class OpenUrlParams(BaseParams):
    url: str

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> None:
        open_url(self.url)


@_register_method('assist/reveal_folder')
class RevealFolderParams(BaseParams):
    path_parts: list[str]

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> None:
        reveal_folder(os.path.join(*self.path_parts))


class SelectFolderResult(TypedDict):
    selection: Path | None


@_register_method('assist/select_folder')
class SelectFolderParams(BaseParams):
    initial_folder: str | None

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> SelectFolderResult:
        app = _toga_handle.get()

        future = concurrent.futures.Future[Path | None]()

        def select_folder():
            async def inner() -> Path | None:
                try:
                    return await app.main_window.select_folder_dialog(  # pyright: ignore[reportUnknownMemberType]
                        'Select folder', self.initial_folder
                    )
                except ValueError:
                    return None

            task = asyncio.create_task(inner())
            task.add_done_callback(lambda t: future.set_result(t.result()))

        app.loop.call_soon_threadsafe(select_folder)
        return {'selection': future.result()}


class ConfirmDialogueResult(TypedDict):
    ok: bool


@_register_method('assist/confirm')
class ConfirmDialogueParams(BaseParams):
    title: str
    message: str

    async def respond(self, config_ctxs: _ConfigBoundCtxCollection) -> ConfirmDialogueResult:
        app = _toga_handle.get()

        future = concurrent.futures.Future[bool]()

        def confirm():
            async def inner():
                return await app.main_window.confirm_dialog(self.title, self.message)

            task = asyncio.create_task(inner())
            task.add_done_callback(lambda t: future.set_result(t.result()))

        app.loop.call_soon_threadsafe(confirm)
        return {'ok': future.result()}


class _ConfigBoundCtxCollection:
    def __init__(self):
        self.locks = WeakValueDefaultDictionary[object, asyncio.Lock](asyncio.Lock)

        self._config_ctxs = dict[str, shared_ctx.ConfigBoundCtx]()

        self.current_progress: ReadOnlyProgressGroup[pkg_management.PkgDownloadProgress] = {}

        self._github_auth_device_codes = None
        self._github_auth_flow_task = None

    async def __aenter__(self):
        self._exit_stack = AsyncExitStack()

        self.global_config = await _read_global_config()

        self._web_client = await self._exit_stack.enter_async_context(
            init_web_client(self.global_config.http_cache_dir, with_progress=True)
        )

        iter_progress = self._exit_stack.enter_context(
            make_progress_receiver[pkg_management.PkgDownloadProgress]()
        )

        async def update_progress():
            async for progress_group in iter_progress:
                self.current_progress = progress_group

        progress_updater = asyncio.create_task(update_progress())
        self._exit_stack.push_async_callback(partial(cancel_tasks, [progress_updater]))

        self._exit_stack.push_async_callback(self.cancel_github_auth_polling)

        shared_ctx.locks_var.set(self.locks)
        shared_ctx.web_client_var.set(self._web_client)

    async def __aexit__(self, *args: object):
        await self._exit_stack.aclose()

    async def load(self, profile: str) -> shared_ctx.ConfigBoundCtx:
        try:
            config_ctx = self._config_ctxs[profile]
        except KeyError:
            async with self.locks[(*_LockOperation.ModifyProfile, profile)]:
                try:
                    config_ctx = self._config_ctxs[profile]
                except KeyError:
                    config_ctx = self._config_ctxs[profile] = shared_ctx.ConfigBoundCtx(
                        await _read_profile_config(self.global_config, profile)
                    )

        return config_ctx

    async def unload(self, profile: str) -> None:
        if profile in self._config_ctxs:
            config_ctx = self._config_ctxs[profile]
            await run_in_thread(config_ctx.__exit__)(*((None,) * 3))
            del self._config_ctxs[profile]

    async def _unload_all(self):
        await gather(self.unload(p) for p in self._config_ctxs)

    async def update_global_config(
        self, update_cb: Callable[[GlobalConfig], GlobalConfig]
    ) -> GlobalConfig:
        async with self.locks[_LockOperation.UpdateGlobalConfig]:
            with _reraise_validation_errors(_ConfigError):
                self.global_config = update_cb(self.global_config)
                await run_in_thread(self.global_config.write)()

            await self._unload_all()

            return self.global_config

    async def initiate_github_auth_flow(self):
        async with self.locks[_LockOperation.InitiateGithubAuthFlow]:
            if self._github_auth_device_codes is None:

                async def finalise_github_auth_flow():
                    result = await poll_for_access_token(
                        self._web_client, codes['device_code'], codes['interval']
                    )

                    def update_global_config_cb(global_config: GlobalConfig):
                        return attrs.evolve(
                            global_config,
                            access_tokens=attrs.evolve(
                                global_config.access_tokens, github=SecretStr(result)
                            ),
                        )

                    await self.update_global_config(update_global_config_cb)

                def on_task_complete(task: object):
                    self._github_auth_flow_task = None
                    self._github_auth_device_codes = None

                self._github_auth_device_codes = codes = await get_codes(self._web_client)
                self._github_auth_flow_task = asyncio.create_task(finalise_github_auth_flow())
                self._github_auth_flow_task.add_done_callback(on_task_complete)

        return self._github_auth_device_codes

    async def wait_for_github_auth_completion(self):
        if self._github_auth_flow_task is not None:
            try:
                await self._github_auth_flow_task
            except BaseException:
                return 'failure'
        return 'success'

    async def cancel_github_auth_polling(self):
        if self._github_auth_flow_task:
            await cancel_tasks([self._github_auth_flow_task])


async def create_web_app(toga_handle: toga.App | None = None):
    if toga_handle:
        _toga_handle.set(toga_handle)

    frontend_resources = importlib.resources.files(frontend)

    config_ctxs = _ConfigBoundCtxCollection()

    async def config_ctxs_listen(app: aiohttp.web.Application):
        async with config_ctxs:
            yield

    async def get_index(request: aiohttp.web.Request):
        return aiohttp.web.Response(
            content_type='text/html',
            text=frontend_resources.joinpath('index.html').read_text(),
        )

    async def get_static_file(request: aiohttp.web.Request):
        filename = request.path.lstrip('/')

        if filename.endswith('.js'):
            content_type = 'application/javascript'
        elif filename.endswith('.js.map'):
            content_type = 'application/json'
        elif filename.endswith('.css'):
            content_type = 'text/css'
        else:
            raise aiohttp.web.HTTPNotFound

        return aiohttp.web.Response(
            content_type=content_type,
            text=frontend_resources.joinpath(filename).read_text(),
        )

    def json_serialize(value: dict[str, Any]):
        return json.dumps(_converter.unstructure(value))

    rpc_server = WsJsonRpcServer(
        json_serialize=json_serialize,
        middlewares=rpc_middlewares.DEFAULT_MIDDLEWARES,  # pyright: ignore[reportUnknownMemberType]
    )
    rpc_server.add_methods(  # pyright: ignore[reportUnknownMemberType]
        [m.bind(n, config_ctxs) for n, m in _methods]
    )

    @aiohttp.web.middleware
    async def enforce_same_origin(
        request: aiohttp.web.Request,
        handler: aiohttp.typedefs.Handler,
    ):
        if request.remote != LOCALHOST:
            raise aiohttp.web.HTTPUnauthorized

        if request.path.startswith('/api'):
            origin = request.headers.get(aiohttp.hdrs.ORIGIN)
            if origin is None:
                raise aiohttp.web.HTTPUnauthorized

            origin_url = URL(origin)
            if (
                origin_url.scheme != request.url.scheme
                or origin_url.host != request.url.host
                or origin_url.port != request.url.port
            ):
                raise aiohttp.web.HTTPUnauthorized

        return await handler(request)

    app = aiohttp.web.Application(middlewares=[enforce_same_origin])
    app.add_routes(
        [
            aiohttp.web.get('/', get_index),
            aiohttp.web.get(
                r'/assets/{name}{extension:(?:\.css|\.js(?:\.map)?)}', get_static_file
            ),
            aiohttp.web.get('/api', rpc_server.handle_http_request),
        ]
    )
    app.cleanup_ctx.append(config_ctxs_listen)
    app.freeze()
    return app


@asynccontextmanager
async def run_web_app(app: aiohttp.web.Application):
    "Fire up the server."
    app_runner = aiohttp.web.AppRunner(app)
    await app_runner.setup()
    assert app_runner.server  # Server is created during setup
    # By omitting the port ``loop.create_server`` will find an available port
    # to bind to - equivalent to creating a socket on port 0.
    server = await asyncio.get_running_loop().create_server(app_runner.server, LOCALHOST)
    host, port = server.sockets[0].getsockname()

    try:
        await server.start_serving()

        server_url = URL.build(scheme='http', host=host, port=port)
        logger.debug(f'JSON-RPC server running on {server_url}')
        yield server_url
    finally:
        # Fark knows how you're supposed to gracefully stop the server:
        #   https://github.com/aio-libs/aiohttp/issues/2950
        await app_runner.shutdown()



from __future__ import annotations

import pytest
from yarl import URL

from instawow._sources.wowi import WowiResolver
from instawow.definitions import Defn
from instawow.pkg_models import Pkg
from instawow.shared_ctx import ConfigBoundCtx


@pytest.fixture
def wowi_resolver(
    iw_config_ctx: ConfigBoundCtx,
):
    return WowiResolver(iw_config_ctx.config)


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_resolve_addon(
    wowi_resolver: WowiResolver,
):
    defn = Defn('wowi', '13188-molinari')

    result = (await wowi_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_changelog_url_format(
    wowi_resolver: WowiResolver,
):
    defn = Defn('wowi', '13188-molinari')

    result = (await wowi_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert result.changelog_url.startswith('data:,')


@pytest.mark.parametrize(
    ('url', 'extracted_alias'),
    [
        (
            'https://www.wowinterface.com/downloads/landing.php?fileid=13188',
            '13188',
        ),
        (
            'https://wowinterface.com/downloads/landing.php?fileid=13188',
            '13188',
        ),
        (
            'https://www.wowinterface.com/downloads/fileinfo.php?id=13188',
            '13188',
        ),
        (
            'https://wowinterface.com/downloads/fileinfo.php?id=13188',
            '13188',
        ),
        (
            'https://www.wowinterface.com/downloads/download13188-Molinari',
            '13188',
        ),
        (
            'https://wowinterface.com/downloads/download13188-Molinari',
            '13188',
        ),
        (
            'https://www.wowinterface.com/downloads/info13188-Molinari.html',
            '13188',
        ),
        (
            'https://wowinterface.com/downloads/info13188-Molinari.html',
            '13188',
        ),
        (
            'https://www.wowinterface.com/downloads/info13188',
            '13188',
        ),
        (
            'https://wowinterface.com/downloads/info13188',
            '13188',
        ),
    ],
)
def test_can_extract_alias_from_url(
    wowi_resolver: WowiResolver,
    url: str,
    extracted_alias: str,
):
    assert wowi_resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

from . import Connection, Row
from ._models import Pkg, make_db_converter


def build_pkg_from_row_mapping(connection: Connection, row_mapping: Row) -> Pkg:
    fk = {'pkg_source': row_mapping['source'], 'pkg_id': row_mapping['id']}
    return make_db_converter().structure(
        {
            **row_mapping,
            'options': connection.execute(
                """
                SELECT any_flavour, any_release_type, version_eq
                FROM pkg_options
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchone(),
            'folders': connection.execute(
                """
                SELECT name
                FROM pkg_folder
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchall(),
            'deps': connection.execute(
                """
                SELECT id
                FROM pkg_dep
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchall(),
            'logged_versions': connection.execute(
                """
                SELECT version, install_time
                FROM pkg_version_log
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                ORDER BY install_time DESC
                LIMIT 10
                """,
                fk,
            ).fetchall(),
        },
        Pkg,
    )


def insert_pkg(pkg: Pkg, transaction: Connection) -> None:
    pkg_values = make_db_converter().unstructure(pkg)

    transaction.execute(
        """
        INSERT INTO pkg (
            source,
            id,
            slug,
            name,
            description,
            url,
            download_url,
            date_published,
            version,
            changelog_url
        )
        VALUES (
            :source,
            :id,
            :slug,
            :name,
            :description,
            :url,
            :download_url,
            :date_published,
            :version,
            :changelog_url
        )
        """,
        pkg_values,
    )
    transaction.execute(
        """
        INSERT INTO pkg_options (
            any_flavour,
            any_release_type,
            version_eq,
            pkg_source,
            pkg_id
        )
        VALUES (
            :any_flavour,
            :any_release_type,
            :version_eq,
            :pkg_source,
            :pkg_id
        )
        """,
        pkg_values['options'] | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']},
    )
    transaction.executemany(
        """
        INSERT INTO pkg_folder (
            name,
            pkg_source,
            pkg_id
        )
        VALUES (
            :name,
            :pkg_source,
            :pkg_id
        )
        """,
        [
            f | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']}
            for f in pkg_values['folders']
        ],
    )
    if pkg_values['deps']:
        transaction.executemany(
            """
            INSERT INTO pkg_dep (
                id,
                pkg_source,
                pkg_id
            )
            VALUES (
                :id,
                :pkg_source,
                :pkg_id
            )
            """,
            [
                f | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']}
                for f in pkg_values['deps']
            ],
        )
    transaction.execute(
        """
        INSERT OR IGNORE INTO pkg_version_log (
            version,
            pkg_source,
            pkg_id
        )
        VALUES (
            :version,
            :pkg_source,
            :pkg_id
        )
        """,
        {
            'version': pkg_values['version'],
            'pkg_source': pkg_values['source'],
            'pkg_id': pkg_values['id'],
        },
    )


def delete_pkg(pkg: Pkg, transaction: Connection) -> None:
    transaction.execute(
        'DELETE FROM pkg WHERE source = :source AND id = :id',
        make_db_converter().unstructure(pkg),
    )



from __future__ import annotations

from collections.abc import Awaitable, Callable, Collection, Iterable, Mapping, Sequence
from contextlib import asynccontextmanager
from functools import wraps
from itertools import chain, compress, filterfalse, product, repeat, starmap
from pathlib import Path
from shutil import move
from tempfile import NamedTemporaryFile
from typing import Concatenate, Literal, TypeVar

import attrs
from typing_extensions import Never, ParamSpec
from yarl import URL

from . import http, pkg_models, shared_ctx
from . import results as R
from ._logging import logger
from ._progress_reporting import Progress, make_incrementing_progress_tracker
from ._utils.aio import gather, run_in_thread
from ._utils.file import trash
from ._utils.iteration import bucketise, uniq
from ._utils.perf import time_op
from ._utils.text import shasum
from ._utils.web import file_uri_to_path, is_file_uri
from .definitions import Defn, Strategy
from .pkg_db import _ops as pkg_db_ops
from .resolvers import HeadersIntent

_T = TypeVar('_T')
_P = ParamSpec('_P')

_AsyncNamedTemporaryFile = run_in_thread(NamedTemporaryFile)
_move_async = run_in_thread(move)

_MUTATE_PKGS_LOCK = '_MUTATE_PKGS_'
_DOWNLOAD_PKG_LOCK = '_DOWNLOAD_PKG_'


class PkgDownloadProgress(Progress[Literal['pkg_download'], Literal['bytes']]):
    profile: str
    defn: Defn


def bucketise_results(
    values: Iterable[tuple[Defn, R.AnyResult[_T]]],
):
    ts: dict[Defn, _T] = {}
    errors: dict[Defn, R.AnyResult[Never]] = {}

    for defn, value in values:
        if isinstance(value, R.ManagerError | R.InternalError):
            errors[defn] = value
        else:
            ts[defn] = value

    return ts, errors


@asynccontextmanager
async def _open_temp_writer_async():
    fh = await _AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, run_in_thread(fh.write))
    except BaseException:
        await run_in_thread(fh.close)()
        await run_in_thread(path.unlink)()
        raise
    else:
        await run_in_thread(fh.close)()


@object.__new__
class _DummyResolver:
    async def resolve(self, defns: Sequence[Defn]):
        return dict.fromkeys(defns, R.PkgSourceInvalid())

    async def get_changelog(self, uri: URL):
        raise R.PkgSourceInvalid


def _with_lock(lock_name: str):
    def outer(
        coro_fn: Callable[Concatenate[shared_ctx.ConfigBoundCtx, _P], Awaitable[_T]],
    ) -> Callable[Concatenate[shared_ctx.ConfigBoundCtx, _P], Awaitable[_T]]:
        @wraps(coro_fn)
        async def inner(
            config_ctx: shared_ctx.ConfigBoundCtx, *args: _P.args, **kwargs: _P.kwargs
        ) -> _T:
            async with shared_ctx.locks[lock_name, config_ctx.config.profile]:
                return await coro_fn(config_ctx, *args, **kwargs)

        return inner

    return outer


def get_alias_from_url(
    config_ctx: shared_ctx.ConfigBoundCtx, value: str
) -> tuple[str, str] | None:
    "Attempt to extract a valid ``Defn`` source and alias from a URL."
    url = URL(value)
    return next(
        (
            (r.metadata.id, a)
            for r in config_ctx.resolvers.values()
            if (a := r.get_alias_from_url(url))
        ),
        None,
    )


def _check_pkgs_not_exist(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Collection[Defn]
) -> list[bool]:
    "Check that packages exist in the database."
    if not defns:
        return []

    with (
        config_ctx.database.connect() as connection,
        config_ctx.database.use_tuple_factory(connection) as cursor,
    ):
        return [
            e
            for (e,) in cursor.execute(
                f"""
                WITH defn (source, alias, id)
                AS (
                    VALUES {", ".join(("(?, ?, ?)",) * len(defns))}
                )
                SELECT NOT EXISTS (
                    SELECT 1
                    FROM pkg
                    WHERE pkg.source = defn.source AND (
                        pkg.id = defn.alias OR pkg.id = defn.id OR lower(pkg.slug) = lower(defn.alias)
                    )
                )
                FROM defn
                """,
                tuple(i for d in defns for i in (d.source, d.alias, d.id)),
            ).fetchall()
        ]


def get_pkgs(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Collection[Defn] | Literal['all']
) -> list[pkg_models.Pkg | None]:
    if defns != 'all' and not defns:
        return []

    with config_ctx.database.connect() as connection:
        if defns == 'all':
            pkgs = connection.execute(
                """
                SELECT * FROM pkg
                """,
            ).fetchall()
        else:
            pkgs = connection.execute(
                f"""
                WITH defn (source, alias, id)
                AS (
                    VALUES {", ".join(("(?, ?, ?)",) * len(defns))}
                )
                SELECT pkg.*
                FROM defn
                LEFT JOIN pkg ON pkg.source = defn.source AND (
                    pkg.id = defn.alias OR pkg.id = defn.id OR lower(pkg.slug) = lower(defn.alias)
                )
                """,
                tuple(i for d in defns for i in (d.source, d.alias, d.id)),
            ).fetchall()

        return [
            pkg_models.build_pkg_from_row_mapping(connection, m) if m['source'] else None
            for m in pkgs
        ]


def get_pkg(config_ctx: shared_ctx.ConfigBoundCtx, defn: Defn) -> pkg_models.Pkg | None:
    "Retrieve a package from the database."
    (pkg,) = get_pkgs(config_ctx, [defn])
    return pkg


async def find_equivalent_pkg_defns(
    config_ctx: shared_ctx.ConfigBoundCtx, pkgs: Collection[pkg_models.Pkg]
) -> dict[pkg_models.Pkg, list[Defn]]:
    "Given a list of packages, find ``Defn``s of each package from other sources."
    from .catalogue import synchronise as synchronise_catalogue
    from .matchers import AddonFolder

    catalogue = await synchronise_catalogue()

    @run_in_thread
    def collect_addon_folders():
        return {
            p: frozenset(
                a
                for f in p.folders
                for a in (
                    AddonFolder.from_path(
                        config_ctx.config.game_flavour,
                        config_ctx.config.addon_dir / f.name,
                    ),
                )
                if a
            )
            for p in pkgs
        }

    async def collect_hashed_folder_defns():
        matches = [
            m
            for r in config_ctx.resolvers.values()
            for m in await r.get_folder_hash_matches(
                [a for p, f in folders_per_pkg.items() if p.source != r.metadata.id for a in f]
            )
        ]

        defns_per_pkg: dict[pkg_models.Pkg, frozenset[Defn]] = {
            p: frozenset() for p in folders_per_pkg
        }
        for (pkg, orig_folders), (defn, matched_folders) in product(
            folders_per_pkg.items(), matches
        ):
            if orig_folders == matched_folders:
                defns_per_pkg[pkg] |= frozenset((defn,))

        return defns_per_pkg

    def get_catalogue_defns(pkg: pkg_models.Pkg) -> frozenset[Defn]:
        entry = catalogue.keyed_entries.get((pkg.source, pkg.id))
        if entry:
            return frozenset(Defn(s.source, s.id) for s in entry.same_as)
        else:
            return frozenset()

    def get_addon_toc_defns(pkg_source: str, addon_folders: Collection[AddonFolder]):
        return frozenset(
            d
            for a in addon_folders
            for d in a.get_defns_from_toc_keys(config_ctx.resolvers.addon_toc_key_and_id_pairs)
            if d.source != pkg_source
        )

    folders_per_pkg = await collect_addon_folders()

    with time_op(lambda t: logger.debug(f'hashed folder matches found in {t:.3f}s')):
        hashed_folder_defns = await collect_hashed_folder_defns()

    return {
        p: sorted(d, key=lambda d: config_ctx.resolvers.priority_dict[d.source])
        for p in pkgs
        for d in (
            get_catalogue_defns(p)
            | get_addon_toc_defns(p.source, folders_per_pkg[p])
            | hashed_folder_defns[p],
        )
        if d
    }


async def _resolve_deps(
    config_ctx: shared_ctx.ConfigBoundCtx, results: Collection[R.AnyResult[pkg_models.Pkg]]
) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
    """Resolve package dependencies.

    The resolver will not follow dependencies
    more than one level deep.  This is to avoid unnecessary
    complexity for something that I would never expect to
    encounter in the wild.
    """
    pkgs = [r for r in results if isinstance(r, pkg_models.Pkg)]
    dep_defns = uniq(
        filterfalse(
            {(p.source, p.id) for p in pkgs}.__contains__,
            ((p.source, d.id) for p in pkgs for d in p.deps),
        )
    )
    if not dep_defns:
        return {}

    deps = await resolve(config_ctx, list(starmap(Defn, dep_defns)))
    pretty_deps = {
        attrs.evolve(d, alias=r.slug) if isinstance(r, pkg_models.Pkg) else d: r
        for d, r in deps.items()
    }
    return pretty_deps


async def resolve(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Collection[Defn], with_deps: bool = False
) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
    "Resolve definitions into packages."
    if not defns:
        return {}

    defns_by_source = bucketise(defns, key=lambda v: v.source)

    track_progress = make_incrementing_progress_tracker(len(defns_by_source), 'Resolving add-ons')

    results = await gather(
        track_progress(
            R.resultify_async_exc(config_ctx.resolvers.get(s, _DummyResolver).resolve(b))
        )
        for s, b in defns_by_source.items()
    )
    results_by_defn = dict(
        chain(
            zip(defns, repeat(R.ManagerError())),
            *(
                r.items() if isinstance(r, dict) else zip(d, repeat(r))
                for d, r in zip(defns_by_source.values(), results)
            ),
        )
    )
    if with_deps:
        results_by_defn.update(await _resolve_deps(config_ctx, results_by_defn.values()))

    return results_by_defn


async def get_changelog(config_ctx: shared_ctx.ConfigBoundCtx, source: str, uri: str) -> str:
    "Retrieve a changelog from a URI."
    return await config_ctx.resolvers.get(source, _DummyResolver).get_changelog(URL(uri))


async def _download_pkg_archive(
    config_ctx: shared_ctx.ConfigBoundCtx, defn: Defn, pkg: pkg_models.Pkg
) -> Path:
    if is_file_uri(pkg.download_url):
        return Path(file_uri_to_path(pkg.download_url))

    async with shared_ctx.locks[_DOWNLOAD_PKG_LOCK, pkg.download_url]:
        headers = await config_ctx.resolvers[pkg.source].make_request_headers(
            intent=HeadersIntent.Download
        )
        trace_request_ctx = {
            'progress': PkgDownloadProgress(
                type_='pkg_download',
                unit='bytes',
                current=0,
                total=0,
                profile=config_ctx.config.profile,
                defn=defn,
            )
        }

        async with (
            shared_ctx.web_client.get(
                pkg.download_url,
                headers=headers,
                raise_for_status=True,
                trace_request_ctx=trace_request_ctx,
                expire_after=http.CACHE_INDEFINITELY,
            ) as response,
            _open_temp_writer_async() as (temp_path, write),
        ):
            async for chunk, _ in response.content.iter_chunks():
                await write(chunk)

        return await _move_async(
            temp_path,
            config_ctx.config.global_config.install_cache_dir / shasum(pkg.download_url),
        )


@run_in_thread
def _install_pkg(
    config_ctx: shared_ctx.ConfigBoundCtx,
    pkg: pkg_models.Pkg,
    archive: Path,
    *,
    replace_folders: bool,
):
    with (
        config_ctx.resolvers.archive_opener_dict[pkg.source](archive) as (
            top_level_folders,
            extract,
        ),
        config_ctx.database.connect() as connection,
    ):
        installed_conflicts = connection.execute(
            f"""
            SELECT DISTINCT pkg.*
            FROM pkg
            JOIN pkg_folder ON pkg_folder.pkg_source = pkg.source AND pkg_folder.pkg_id = pkg.id
            WHERE pkg_folder.name IN ({', '.join(('?',) * len(top_level_folders))})
            """,
            tuple(top_level_folders),
        ).fetchall()
        if installed_conflicts:
            raise R.PkgConflictsWithInstalled(installed_conflicts)

        if replace_folders:
            trash(
                (config_ctx.config.addon_dir / f for f in top_level_folders),
                dest=config_ctx.config.global_config.temp_dir,
                missing_ok=True,
            )
        else:
            unreconciled_conflicts = top_level_folders & {
                f.name for f in config_ctx.config.addon_dir.iterdir()
            }
            if unreconciled_conflicts:
                raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

        extract(config_ctx.config.addon_dir)

        pkg = attrs.evolve(
            pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)]
        )
        with config_ctx.database.transact(connection) as transaction:
            pkg_db_ops.insert_pkg(pkg, transaction)

    return R.PkgInstalled(pkg)


@run_in_thread
def _update_pkg(
    config_ctx: shared_ctx.ConfigBoundCtx,
    old_pkg: pkg_models.Pkg,
    new_pkg: pkg_models.Pkg,
    archive: Path,
):
    with (
        config_ctx.resolvers.archive_opener_dict[new_pkg.source](archive) as (
            top_level_folders,
            extract,
        ),
        config_ctx.database.connect() as connection,
    ):
        installed_conflicts = connection.execute(
            f"""
            SELECT DISTINCT pkg.*
            FROM pkg
            JOIN pkg_folder ON pkg_folder.pkg_source = pkg.source AND pkg_folder.pkg_id = pkg.id
            WHERE (pkg_folder.pkg_source != ? AND pkg_folder.pkg_id != ?)
                AND (pkg_folder.name IN ({', '.join(('?',) * len(top_level_folders))}))
            """,
            (new_pkg.source, new_pkg.id, *top_level_folders),
        ).fetchall()
        if installed_conflicts:
            raise R.PkgConflictsWithInstalled(installed_conflicts)

        unreconciled_conflicts = top_level_folders - {f.name for f in old_pkg.folders} & {
            f.name for f in config_ctx.config.addon_dir.iterdir()
        }
        if unreconciled_conflicts:
            raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

        trash(
            (config_ctx.config.addon_dir / f.name for f in old_pkg.folders),
            dest=config_ctx.config.global_config.temp_dir,
            missing_ok=True,
        )
        extract(config_ctx.config.addon_dir)

        new_pkg = attrs.evolve(
            new_pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)]
        )
        with config_ctx.database.transact(connection) as transaction:
            pkg_db_ops.delete_pkg(old_pkg, transaction)
            pkg_db_ops.insert_pkg(new_pkg, transaction)

    return R.PkgUpdated(old_pkg, new_pkg)


@run_in_thread
def _remove_pkg(config_ctx: shared_ctx.ConfigBoundCtx, pkg: pkg_models.Pkg, *, keep_folders: bool):
    if not keep_folders:
        trash(
            (config_ctx.config.addon_dir / f.name for f in pkg.folders),
            dest=config_ctx.config.global_config.temp_dir,
            missing_ok=True,
        )

    with (
        config_ctx.database.connect() as connection,
        config_ctx.database.transact(connection) as transaction,
    ):
        pkg_db_ops.delete_pkg(pkg, transaction)

    return R.PkgRemoved(pkg)


@run_in_thread
def _check_installed_pkg_integrity(config_ctx: shared_ctx.ConfigBoundCtx, pkg: pkg_models.Pkg):
    return all((config_ctx.config.addon_dir / p.name).exists() for p in pkg.folders)


@_with_lock(_MUTATE_PKGS_LOCK)
async def install(
    config_ctx: shared_ctx.ConfigBoundCtx,
    defns: Sequence[Defn],
    *,
    replace_folders: bool,
    dry_run: bool = False,
) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
    "Install packages from a definition list."

    # We'll weed out installed deps from the results after resolving -
    # doing it this way isn't particularly efficient but avoids having to
    # deal with local state in ``resolve``.
    resolve_results = await resolve(
        config_ctx, list(compress(defns, _check_pkgs_not_exist(config_ctx, defns))), with_deps=True
    )
    pkgs, resolve_errors = bucketise_results(resolve_results.items())
    new_pkgs = dict(
        compress(
            pkgs.items(), _check_pkgs_not_exist(config_ctx, [p.to_defn() for p in pkgs.values()])
        )
    )

    results = dict.fromkeys(defns, R.PkgAlreadyInstalled()) | resolve_errors

    if dry_run:
        return results | {d: R.PkgInstalled(p, dry_run=True) for d, p in new_pkgs.items()}

    download_results = await gather(
        R.resultify_async_exc(_download_pkg_archive(config_ctx, d, r)) for d, r in new_pkgs.items()
    )
    archive_paths, download_errors = bucketise_results(zip(new_pkgs, download_results))

    track_progress = make_incrementing_progress_tracker(len(archive_paths), 'Installing')

    return (
        results
        | download_errors
        | {
            d: await track_progress(
                R.resultify_async_exc(
                    _install_pkg(config_ctx, new_pkgs[d], a, replace_folders=replace_folders)
                )
            )
            for d, a in archive_paths.items()
        }
    )


@_with_lock(_MUTATE_PKGS_LOCK)
async def replace(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Mapping[Defn, Defn]
) -> Mapping[Defn, R.AnyResult[R.PkgInstalled | R.PkgRemoved]]:
    "Replace installed packages with re-reconciled packages."

    inverse_defns = {v: k for k, v in defns.items()}
    if len(inverse_defns) != len(defns):
        raise ValueError('``defns`` must be unique')

    old_pkgs = {d: p for d, p in zip(defns, get_pkgs(config_ctx, defns)) if p}

    resolve_results = await resolve(
        config_ctx,
        list(compress(defns.values(), _check_pkgs_not_exist(config_ctx, defns.values()))),
        with_deps=True,
    )
    pkgs, resolve_errors = bucketise_results(resolve_results.items())
    new_pkgs = dict(
        compress(
            pkgs.items(), _check_pkgs_not_exist(config_ctx, [p.to_defn() for p in pkgs.values()])
        )
    )

    results = dict(
        chain.from_iterable(
            zip(
                ((d, R.PkgNotInstalled()) for d in defns),
                ((d, None) for d in defns.values()),
                strict=True,
            )
        )
    )
    results = results | resolve_errors

    download_results = await gather(
        R.resultify_async_exc(_download_pkg_archive(config_ctx, d, r)) for d, r in new_pkgs.items()
    )
    archive_paths, download_errors = bucketise_results(zip(new_pkgs, download_results))

    replace_coros = {
        k: R.resultify_async_exc(c)
        for d, a in archive_paths.items()
        for o in (inverse_defns[d],)
        for k, c in (
            (o, _remove_pkg(config_ctx, old_pkgs[o], keep_folders=False)),
            (d, _install_pkg(config_ctx, new_pkgs[d], a, replace_folders=False)),
        )
    }

    track_progress = make_incrementing_progress_tracker(len(replace_coros), 'Replacing')

    results = (
        results | download_errors | {k: await track_progress(c) for k, c in replace_coros.items()}
    )
    return {k: v for k, v in results.items() if v is not None}


@_with_lock(_MUTATE_PKGS_LOCK)
async def update(
    config_ctx: shared_ctx.ConfigBoundCtx,
    defns: Sequence[Defn] | Literal['all'],
    *,
    dry_run: bool = False,
) -> Mapping[Defn, R.AnyResult[R.PkgInstalled | R.PkgUpdated]]:
    "Update installed packages from a definition list."

    if defns == 'all':
        defns_to_pkgs = {p.to_defn(): p for p in get_pkgs(config_ctx, defns) if p}
        defns = list(defns_to_pkgs)

        resolve_defns = {d: d for d in defns_to_pkgs}

    else:
        defns_to_pkgs = {d: p for d, p in zip(defns, get_pkgs(config_ctx, defns)) if p}

        resolve_defns = {
            # Attach the source ID to each ``Defn`` from the
            # corresponding installed package.  Using the ID has the benefit
            # of resolving installed-but-renamed packages - the slug is
            # transient but the ID isn't
            attrs.evolve(d, id=p.id) if d.strategies.initialised else p.to_defn(): d
            for d, p in defns_to_pkgs.items()
        }

    resolve_results = await resolve(config_ctx, resolve_defns, with_deps=True)
    pkgs, resolve_errors = bucketise_results(
        # Discard the reconstructed ``Defn``s
        (resolve_defns.get(d) or d, r)
        for d, r in resolve_results.items()
    )

    updatables = {
        d: (o, n)
        for d, n in pkgs.items()
        for o in (defns_to_pkgs.get(d),)
        if not o
        or o.version != n.version
        or not await _check_installed_pkg_integrity(config_ctx, o)
    }

    results = (
        dict.fromkeys(defns, R.PkgNotInstalled())
        | resolve_errors
        | {
            d: R.PkgUpToDate(is_pinned=pkgs[d].options.version_eq)
            for d in pkgs.keys() - updatables.keys()
        }
    )

    if dry_run:
        return results | {
            d: R.PkgUpdated(o, n, dry_run=True) if o else R.PkgInstalled(n, dry_run=True)
            for d, (o, n) in updatables.items()
        }

    download_results = await gather(
        R.resultify_async_exc(_download_pkg_archive(config_ctx, d, n))
        for d, (_, n) in updatables.items()
    )
    archive_paths, download_errors = bucketise_results(zip(updatables, download_results))

    track_progress = make_incrementing_progress_tracker(len(archive_paths), 'Updating')

    return (
        results
        | download_errors
        | {
            d: await track_progress(
                R.resultify_async_exc(
                    _update_pkg(config_ctx, o, n, a)
                    if o
                    else _install_pkg(config_ctx, n, a, replace_folders=False)
                )
            )
            for d, a in archive_paths.items()
            for o, n in (updatables[d],)
        }
    )


@_with_lock(_MUTATE_PKGS_LOCK)
async def remove(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Sequence[Defn], *, keep_folders: bool
) -> Mapping[Defn, R.AnyResult[R.PkgRemoved]]:
    "Remove packages by their definition."
    return {
        d: (
            await R.resultify_async_exc(_remove_pkg(config_ctx, p, keep_folders=keep_folders))
            if p
            else R.PkgNotInstalled()
        )
        for d, p in zip(defns, get_pkgs(config_ctx, defns))
    }


@_with_lock(_MUTATE_PKGS_LOCK)
async def pin(
    config_ctx: shared_ctx.ConfigBoundCtx, defns: Sequence[Defn]
) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
    """Pin and unpin installed packages.

    instawow does not have true pinning.  This flips ``Strategy.VersionEq``
    on for installed packages from sources that support it.
    The net effect is the same as if the package
    had been reinstalled with the ``VersionEq`` strategy.
    """

    async def pin(defn: Defn, pkg: pkg_models.Pkg | None) -> R.PkgInstalled:
        resolver = config_ctx.resolvers.get(defn.source)
        if resolver is None:
            raise R.PkgSourceInvalid
        elif Strategy.VersionEq not in resolver.metadata.strategies:
            raise R.PkgStrategiesUnsupported({Strategy.VersionEq})

        if not pkg:
            raise R.PkgNotInstalled

        version = defn.strategies[Strategy.VersionEq]
        if version and pkg.version != version:
            R.PkgFilesNotMatching(defn.strategies)

        version_eq = version is not None

        with (
            config_ctx.database.connect() as connection,
            config_ctx.database.transact(connection) as transaction,
        ):
            transaction.execute(
                """
                UPDATE pkg_options
                SET version_eq = :version_eq
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                {
                    'pkg_source': pkg.source,
                    'pkg_id': pkg.id,
                    'version_eq': version_eq,
                },
            )

        return R.PkgInstalled(
            attrs.evolve(pkg, options=attrs.evolve(pkg.options, version_eq=version_eq)),
        )

    return {
        d: await R.resultify_async_exc(pin(d, p))
        for d, p in zip(defns, get_pkgs(config_ctx, defns))
    }



from __future__ import annotations

import time
from collections.abc import Callable, Iterator
from contextlib import contextmanager


@contextmanager
def time_op(on_complete: Callable[[float], None]) -> Iterator[None]:
    start = time.perf_counter()
    yield
    on_complete(time.perf_counter() - start)



from __future__ import annotations

import multiprocessing
import sys
from pathlib import Path

from loguru import logger


def _intercept_logging_module_calls(log_level: str):  # pragma: no cover
    import inspect
    import logging

    logging_filename = getattr(logging, '__file__', None)

    class InterceptHandler(logging.Handler):
        def emit(self, record: logging.LogRecord) -> None:
            # Get the corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where the logged message originated
            depth = 0
            frame = inspect.currentframe()
            while frame and (depth == 0 or frame.f_code.co_filename == logging_filename):
                frame = frame.f_back
                depth += 1

            logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

    logging.basicConfig(handlers=[InterceptHandler()], level=log_level, force=True)


def setup_logging(
    logging_dir: Path, log_to_stderr: bool, debug: bool, intercept_logging_module_calls: bool
) -> None:
    log_level = 'DEBUG' if debug else 'INFO'

    if intercept_logging_module_calls:
        _intercept_logging_module_calls(log_level)

    context = None
    if sys.platform == 'darwin' and 'fork' in multiprocessing.get_all_start_methods():
        # Avoid the overhead of starting the "spawn" method's resource tracker.
        # instawow doesn't use multi-processing at all and neither does loguru;
        # it simply creates MP sync primitives to support applications that do.
        # When `enqueue` is `True`, messages are logged in a sub-thread of the
        # main process and not in a separate process.
        context = multiprocessing.get_context('fork')

    handlers = [
        {
            'level': log_level,
            'enqueue': True,
            'context': context,
            'sink': logging_dir / 'error.log',
            'rotation': '5 MB',
            'retention': 5,  # Number of log files to keep
        },
    ]
    if log_to_stderr:
        handlers += [
            {
                'level': log_level,
                'enqueue': True,
                'context': context,
                'sink': sys.stderr,
            }
        ]

    logger.configure(handlers=handlers)



from __future__ import annotations

from pathlib import Path

import pytest

from instawow.definitions import Defn
from instawow.matchers import (
    AddonFolder,
    AddonHashMethod,
    Matcher,
    get_unreconciled_folders,
    hash_addon_contents,
    match_addon_names_with_folder_names,
    match_folder_hashes,
    match_folder_name_subsets,
    match_toc_source_ids,
)
from instawow.shared_ctx import ConfigBoundCtx
from instawow.wow_installations import Flavour

MOLINARI_HASH = '2da096db5769138b5428a068343cddf3'


def write_addons(
    iw_config_ctx: ConfigBoundCtx,
    *addons: str,
):
    for addon in addons:
        (iw_config_ctx.config.addon_dir / addon).mkdir()
        (iw_config_ctx.config.addon_dir / addon / f'{addon}.toc').touch()


@pytest.fixture
def molinari(
    iw_config_ctx: ConfigBoundCtx,
):
    molinari_folder = iw_config_ctx.config.addon_dir / 'Molinari'
    molinari_folder.mkdir()

    with open(molinari_folder / 'Molinari.toc', 'w', newline='\n') as toc_file:
        toc_file.write(
            """\
## X-Curse-Project-ID: 20338
## X-WoWI-ID: 13188
""",
        )

    return molinari_folder


def test_can_extract_defns_from_addon_folder_toc(
    iw_config_ctx: ConfigBoundCtx,
    molinari: Path,
):
    addon_folder = AddonFolder.from_path(iw_config_ctx.config.game_flavour, molinari)
    assert addon_folder
    assert addon_folder.get_defns_from_toc_keys(
        iw_config_ctx.resolvers.addon_toc_key_and_id_pairs
    ) == {Defn('curse', '20338'), Defn('wowi', '13188')}


def test_addon_folder_is_hashable(
    iw_config_ctx: ConfigBoundCtx,
    molinari: Path,
):
    addon_folder = AddonFolder.from_path(iw_config_ctx.config.game_flavour, molinari)
    assert addon_folder
    assert hash_addon_contents(addon_folder.path, AddonHashMethod.Wowup) == MOLINARI_HASH


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_reconcile_invalid_addons_discarded(
    iw_config_ctx: ConfigBoundCtx,
):
    iw_config_ctx.config.addon_dir.joinpath('foo').mkdir()
    iw_config_ctx.config.addon_dir.joinpath('bar').touch()
    folders = get_unreconciled_folders(iw_config_ctx)
    assert folders == frozenset()
    assert await match_toc_source_ids(iw_config_ctx, folders) == []
    assert await match_folder_name_subsets(iw_config_ctx, folders) == []


@pytest.mark.parametrize(
    ('test_func', 'expected_defns'),
    [
        (
            match_toc_source_ids,
            {
                Defn('curse', '20338'),
                Defn('wowi', '13188'),
                Defn('github', '388670'),
                Defn('wago', 'WqKQQEKx'),
            },
        ),
        (
            match_folder_hashes,
            {
                Defn('wago', 'WqKQQEKx'),
            },
        ),
        (
            match_folder_name_subsets,
            {
                Defn('curse', '20338'),
                Defn('wowi', '13188'),
            },
        ),
        (
            match_addon_names_with_folder_names,
            {
                Defn('curse', '20338'),
                Defn('wowi', '13188'),
                Defn('github', '388670'),
            },
        ),
    ],
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_reconcile_multiple_defns_per_addon_contained_in_results(
    iw_config_ctx: ConfigBoundCtx,
    molinari: Path,
    test_func: Matcher,
    expected_defns: set[Defn],
):
    ((_, matches),) = await test_func(iw_config_ctx, get_unreconciled_folders(iw_config_ctx))
    assert expected_defns == set(matches)


@pytest.mark.parametrize(
    ('iw_profile_config_values', 'expected_defns'),
    [
        (
            Flavour.Retail,
            {
                Defn('curse', '23350'),
                Defn('curse', '333072'),
                Defn('curse', '431557'),
                Defn('curse', '674779'),
                Defn('curse', '912615'),
            },
        ),
        (
            Flavour.Classic,
            {
                Defn('curse', '23350'),
                Defn('curse', '674779'),
                Defn('curse', '912615'),
            },
        ),
        (
            Flavour.VanillaClassic,
            {
                Defn('curse', '23350'),
                Defn('curse', '333072'),
                Defn('curse', '674779'),
                Defn('curse', '912615'),
            },
        ),
    ],
    indirect=['iw_profile_config_values'],
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_reconcile_results_vary_by_game_flavour(
    iw_config_ctx: ConfigBoundCtx,
    expected_defns: set[Defn],
):
    write_addons(iw_config_ctx, 'AdiBags', 'AdiBags_Config')
    ((_, matches),) = await match_folder_name_subsets(
        iw_config_ctx, get_unreconciled_folders(iw_config_ctx)
    )
    assert expected_defns == set(matches)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path
from typing import Any, TypeAlias

import aiohttp

ClientSession: TypeAlias = aiohttp.ClientSession

_USER_AGENT = 'instawow (+https://github.com/layday/instawow)'

_DEFAULT_EXPIRE = 0  # Do not cache by default.
CACHE_INDEFINITELY = -1

_PROGRESS_TICK_INTERVAL = 0.1


@asynccontextmanager
async def _setup_progress_tracker():
    import asyncio

    from .._progress_reporting import get_next_progress_id, update_progress
    from .._utils.aio import cancel_tasks

    progress_tickers = set[asyncio.Task[None]]()

    async def do_on_request_end(
        _client_session: aiohttp.ClientSession,
        trace_config_ctx: Any,
        params: aiohttp.TraceRequestEndParams,
        /,
    ):
        progress = (
            trace_config_ctx.trace_request_ctx and trace_config_ctx.trace_request_ctx['progress']
        )
        if progress:

            async def do_update_progress():
                response = params.response

                progress_id = get_next_progress_id()
                progress['total'] = (
                    None
                    if aiohttp.hdrs.CONTENT_ENCODING in response.headers
                    else response.content_length
                )

                update_progress(progress_id, progress)

                try:
                    while not response.content.is_eof():
                        progress['current'] = response.content.total_bytes
                        update_progress(progress_id, progress)

                        await asyncio.sleep(_PROGRESS_TICK_INTERVAL)

                finally:
                    update_progress(progress_id, 'unset')

            task = asyncio.create_task(do_update_progress())
            task.add_done_callback(progress_tickers.remove)
            progress_tickers.add(task)

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()

    try:
        yield trace_config
    finally:
        await cancel_tasks(progress_tickers)


@asynccontextmanager
async def init_web_client(
    cache_dir: Path | None, *, no_cache: bool = False, with_progress: bool = False, **kwargs: Any
) -> AsyncIterator[ClientSession]:
    import ssl

    import truststore

    kwargs = {
        'connector': aiohttp.TCPConnector(
            limit_per_host=20, ssl=truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ),
        'headers': {'User-Agent': _USER_AGENT},
        'trust_env': True,  # Respect the ``http(s)_proxy`` env var
        'timeout': aiohttp.ClientTimeout(connect=60, sock_connect=10, sock_read=20),
        **kwargs,
    }

    async with AsyncExitStack() as async_exit_stack:
        if with_progress:
            progress_trace_config = await async_exit_stack.enter_async_context(
                _setup_progress_tracker()
            )
            kwargs['trace_configs'] = [*kwargs.get('trace_configs', []), progress_trace_config]

        if cache_dir is not None:
            from aiohttp_client_cache import CacheBackend
            from aiohttp_client_cache.session import CachedSession

            from ._cache import make_cache

            cache_backend = CacheBackend(
                allowed_codes=(200, 206),
                allowed_methods=('GET', 'POST'),
                expire_after=_DEFAULT_EXPIRE,
                include_headers=True,
            )

            cache = await async_exit_stack.enter_async_context(make_cache(cache_dir))
            cache_backend.responses = cache_backend.redirects = cache
            if no_cache:
                cache_backend.disabled = True

            client_session = await async_exit_stack.enter_async_context(
                CachedSession(cache=cache_backend, **kwargs)
            )
            yield client_session

        else:
            client_session = await async_exit_stack.enter_async_context(
                aiohttp.ClientSession(**kwargs)
            )
            yield client_session



# pyright: reportUnusedFunction=false

from __future__ import annotations

import enum
from collections.abc import Sequence
from typing import Any, Generic, Literal, TypeVar, overload

import attrs
import cattrs
from prompt_toolkit.application import Application
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import IsDone
from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    FormattedText,
    StyleAndTextTuples,
    to_formatted_text,
)
from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    HSplit,
    Layout,
    Window,
)
from prompt_toolkit.shortcuts import PromptSession
from prompt_toolkit.shortcuts.progress_bar import ProgressBar as _ProgressBar
from prompt_toolkit.shortcuts.progress_bar import ProgressBarCounter as _ProgressBarCounter
from prompt_toolkit.shortcuts.progress_bar import formatters as pb_formatters
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator
from prompt_toolkit.widgets import Label
from typing_extensions import Never

from ._utils.compat import fauxfrozen
from ._utils.web import open_url

_T = TypeVar('_T')


class AttrsFieldValidator(Validator):
    "One-off validators for attrs fields."

    def __init__(self, attribute: attrs.Attribute[object], converter: cattrs.Converter) -> None:
        self._field_name = attribute.name
        self._FieldWrapper = attrs.make_class(
            '_FieldWrapper',
            {
                self._field_name: attrs.field(
                    default=attribute.default,
                    validator=attribute.validator,
                    repr=attribute.repr,
                    hash=attribute.hash,
                    init=attribute.init,
                    metadata=attribute.metadata,
                    type=attribute.type,
                    converter=attribute.converter,
                    kw_only=attribute.kw_only,
                    eq=attribute.eq,
                    order=attribute.order,
                    on_setattr=attribute.on_setattr,
                )
            },
        )
        self._converter = converter

    def validate(self, document: Document) -> None:
        try:
            self._converter.structure({self._field_name: document.text}, self._FieldWrapper)
        except Exception as exc:
            raise ValidationError(
                0, '\n'.join(cattrs.transform_error(exc, format_exception=lambda e, _: str(e)))
            ) from exc


@fauxfrozen
class Choice(Generic[_T]):
    label: str | StyleAndTextTuples
    value: _T
    disabled: bool = False
    browser_url: str | None = None


class _FauxPromptSession(Generic[_T]):
    def __init__(self, application: Application[_T]) -> None:
        self.application = application

    def prompt(self) -> _T:
        return self.application.run()

    async def prompt_async(self) -> _T:
        return await self.application.run_async()


_style = Style(
    [
        ('indicator', 'fg:ansicyan'),
        ('question', 'bold'),
        ('skipped', 'fg:ansiyellow'),
        ('attention', 'fg:ansimagenta'),
    ]
)


class _Skip(enum.Enum):
    Skip = enum.auto()


SKIP = _Skip.Skip
_skip_choice = Choice([('underline', 's'), ('', 'kip')], SKIP)


def confirm(message: str) -> PromptSession[bool]:
    bindings = KeyBindings()

    result = None

    @bindings.add('y')
    @bindings.add('Y')
    def yes(event: KeyPressEvent) -> None:
        nonlocal result
        result = True
        event.app.exit(result=result)

    @bindings.add('n')
    @bindings.add('N')
    def no(event: KeyPressEvent) -> None:
        nonlocal result
        result = False
        event.app.exit(result=result)

    @bindings.add(Keys.Enter)  # Override <enter> from the prompt session's base bindings
    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent) -> None:
        "Disallow inserting other text."

    def get_messsage():
        tokens = [
            ('class:indicator', '?' if result is None else '✓' if result else '✗'),
            ('', ' '),
            ('class:question', message),
        ]
        if result is None:
            tokens += [('', '  (Y/n)')]
        return FormattedText(tokens)

    session = PromptSession[bool](
        get_messsage,
        key_bindings=bindings,
        style=_style,
    )
    return session


def path(
    message: str, *, only_directories: bool = False, validator: Validator | None = None
) -> PromptSession[str]:
    completer = PathCompleter(
        expanduser=True,
        only_directories=only_directories,
    )
    session = PromptSession[str](
        FormattedText(
            [('class:indicator', '>'), ('', ' '), ('class:question', message), ('', '  ')]
        ),
        completer=completer,
        style=_style,
        validator=validator,
    )
    return session


def password(message: str) -> PromptSession[str]:
    session = PromptSession[str](
        FormattedText(
            [('class:indicator', '>'), ('', ' '), ('class:question', message), ('', '  ')]
        ),
        is_password=True,
        style=_style,
    )
    return session


def select_multiple(
    message: str,
    choices: Sequence[Choice[_T]],
) -> _FauxPromptSession[list[_T]]:
    bindings = KeyBindings()

    answered = False

    position = 0
    selected_indices = set[int]()

    @bindings.add(Keys.ControlC)
    @bindings.add(Keys.SIGINT)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Up)
    def previous_item(event: KeyPressEvent):
        nonlocal position
        position = max(0, position - 1)

    @bindings.add(Keys.Down)
    def next_item(event: KeyPressEvent):
        nonlocal position
        position = min(len(choices) - 1, position + 1)

    @bindings.add(' ')
    def select_item(event: KeyPressEvent):
        nonlocal selected_indices
        selected_indices = selected_indices ^ {position}

    @bindings.add('o')
    @bindings.add('O')
    def open_item_browser_url(event: KeyPressEvent):
        choice_at_position = choices[position]
        if choice_at_position.browser_url:
            open_url(choice_at_position.browser_url)

    @bindings.add(Keys.Enter)
    def submit(event: KeyPressEvent):
        nonlocal answered
        answered = True
        event.app.exit(result=[choices[i].value for i in selected_indices])

    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent):
        "Disallow inserting other text."

    def get_label_messsage():
        tokens = [
            ('class:indicator', '✓' if answered else '-'),
            ('', ' '),
            ('class:question', message),
        ]
        if not answered:
            tokens += [
                (
                    '',
                    '  (use arrow keys to move, <space> to select'
                    + (', <o> to open in browser' if any(c.browser_url for c in choices) else '')
                    + ' and <enter> to confirm)',
                )
            ]
        return FormattedText(tokens)

    def get_select_tokens():
        tokens = list[StyleAndTextTuples]()

        for i, choice in enumerate(choices):
            focussed = i == position
            if focussed:
                tokens += [('[SetCursorPosition]', '')]

            if choice.disabled:
                tokens += [('', ' ')]
            else:
                selected = i in selected_indices
                if selected:
                    tokens += [('class:checkbox-selected', '■')]
                else:
                    tokens += [('', '□')]

            tokens += [('', ' '), *to_formatted_text(choice.label), ('', '\n')]

        return tokens[:-1]

    app = Application[list[_T]](
        key_bindings=bindings,
        layout=Layout(
            HSplit(
                [
                    Label(get_label_messsage),
                    ConditionalContainer(
                        Window(FormattedTextControl(get_select_tokens, focusable=True)),
                        filter=~IsDone(),
                    ),
                ],
            )
        ),
        style=_style,
    )
    return _FauxPromptSession(app)


@overload
def select_one(
    message: str,
    choices: Sequence[Choice[_T]],
    *,
    can_skip: Literal[True],
    initial_value: _T | None = None,
) -> _FauxPromptSession[_T | Literal[SKIP]]: ...
@overload
def select_one(
    message: str,
    choices: Sequence[Choice[_T]],
    *,
    can_skip: Literal[False] = False,
    initial_value: _T | None = None,
) -> _FauxPromptSession[_T]: ...


def select_one(
    message: str,
    choices: Sequence[Choice[_T]],
    *,
    can_skip: bool = False,
    initial_value: _T | None = None,
) -> _FauxPromptSession[Any]:
    bindings = KeyBindings()

    answered = False

    combined_choices = [*choices, _skip_choice] if can_skip else choices
    positions = [i for i, c in enumerate(combined_choices) if not c.disabled]
    position = next(
        p for p in positions if initial_value is None or choices[p].value == initial_value
    )

    @bindings.add(Keys.ControlC)
    @bindings.add(Keys.SIGINT)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Up)
    def previous_item(event: KeyPressEvent):
        nonlocal position
        try:
            position = positions[positions.index(position) - 1]
        except IndexError:
            position = positions[-1]

    @bindings.add(Keys.Down)
    def next_item(event: KeyPressEvent):
        nonlocal position
        try:
            position = positions[positions.index(position) + 1]
        except IndexError:
            position = positions[0]

    @bindings.add('o')
    @bindings.add('O')
    def open_item_browser_url(event: KeyPressEvent):
        choice_at_position = combined_choices[position]
        if choice_at_position.browser_url:
            open_url(choice_at_position.browser_url)

    @bindings.add(Keys.Enter)
    def submit(event: KeyPressEvent):
        nonlocal answered
        answered = True
        event.app.exit(result=combined_choices[position].value)

    if can_skip:

        @bindings.add('s')
        def skip_question(event: KeyPressEvent):
            nonlocal answered, position
            answered = True
            position = positions[-1]
            event.app.exit(result=SKIP)

    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent):
        "Disallow inserting other text."

    def get_label_messsage():
        tokens = [
            (
                'class:indicator',
                '✓' if answered and combined_choices[position] is not _skip_choice else '-',
            ),
            ('', ' '),
            ('class:question', message),
        ]
        if answered and combined_choices[position] is _skip_choice:
            tokens += [('class:skipped', '  (skipped)')]
        elif not answered:
            tokens += [
                (
                    '',
                    '  (use arrow keys to move'
                    + (
                        ', <o> to open in browser'
                        if any(c.browser_url for c in combined_choices)
                        else ''
                    )
                    + ' and <enter> to select)',
                )
            ]
        else:
            tokens += [
                ('', '  ('),
                *to_formatted_text(combined_choices[position].label),
                ('', ')'),
            ]
        return FormattedText(tokens)

    def get_select_tokens():
        tokens = list[StyleAndTextTuples]()

        for i, choice in enumerate(combined_choices):
            if choice.disabled:
                tokens += [('', ' ')]
            else:
                focussed = i == position
                if focussed:
                    tokens += [('[SetCursorPosition]', ''), ('class:radio-selected', '●')]
                else:
                    tokens += [('', '○')]

            tokens += [('', ' '), *to_formatted_text(choice.label), ('', '\n')]

        return tokens[:-1]

    app = Application[Any](
        key_bindings=bindings,
        layout=Layout(
            HSplit(
                [
                    Label(get_label_messsage),
                    ConditionalContainer(
                        Window(FormattedTextControl(get_select_tokens, focusable=True)),
                        filter=~IsDone(),
                    ),
                ],
            )
        ),
        style=_style,
    )
    return _FauxPromptSession(app)


class ProgressBar(_ProgressBarCounter[Never]):
    def __init__(
        self,
        *,
        progress_bar: _ProgressBar,
        is_download: bool,
        label: AnyFormattedText = '',
        total: int | None = None,
    ) -> None:
        super().__init__(progress_bar=progress_bar, label=label, total=total)
        self.is_download = is_download


class _DownloadProgress(pb_formatters.Progress):
    template = '<current>{current:>3.1f}</current>/<total>{total:>3.1f}</total>MB'

    def format(
        self, progress_bar: _ProgressBar, progress: _ProgressBarCounter[object], width: int
    ):
        if isinstance(progress, ProgressBar) and not progress.is_download:
            return super().format(progress_bar, progress, width)

        return HTML(self.template).format(
            current=progress.items_completed / 2**20, total=(progress.total or 0) / 2**20
        )


def make_progress_bar_group() -> _ProgressBar:
    "``ProgressBar`` with download progress expressed in megabytes."
    return _ProgressBar(
        formatters=[
            pb_formatters.Label(),
            pb_formatters.Text(' '),
            pb_formatters.Percentage(),
            pb_formatters.Text(' '),
            pb_formatters.Bar(sym_a='#', sym_b='#', sym_c='.'),
            pb_formatters.Text(' '),
            _DownloadProgress(),
        ],
    )



from __future__ import annotations

import logging
import re
from io import BytesIO
from typing import Literal
from zipfile import ZipFile

import aiohttp.hdrs
import aiohttp.web
import pytest
from aresponses import ResponsesMockServer
from yarl import URL

from instawow._sources.github import GithubResolver
from instawow.definitions import Defn, Strategies, Strategy
from instawow.pkg_models import Pkg
from instawow.results import PkgFilesMissing, PkgFilesNotMatching, PkgNonexistent
from instawow.shared_ctx import ConfigBoundCtx
from instawow.wow_installations import Flavour, FlavourVersionRange

from .fixtures.http import Route


@pytest.fixture
def github_resolver(
    iw_config_ctx: ConfigBoundCtx,
):
    return GithubResolver(iw_config_ctx.config)


zip_defn = Defn('github', '28/NoteworthyII')
zip_addon_name = URL(zip_defn.alias).name

packager_test_defn = Defn('github', 'nebularg/PackagerTest')


ZIPS = {
    'flavoured-toc-only': {
        'toc_files': {
            '_Cata': b'',
        },
        'flavours': {Flavour.Classic},
    },
    'flavoured-and-unflavoured-toc-without-interface-version': {
        'toc_files': {
            '_Cata': b'',
            '': b'',
        },
        'flavours': {Flavour.Classic},
    },
    'flavoured-and-unflavoured-toc-with-interface-version': {
        'toc_files': {
            '_Cata': b'',
            '': b'## Interface: 11300\n',
        },
        'flavours': {Flavour.VanillaClassic, Flavour.Classic},
    },
    'unflavoured-toc-only-without-interface-version': {
        'toc_files': {
            '': b'',
        },
        'flavours': set(),
    },
    'unflavoured-toc-only-with-interface-version': {
        'toc_files': {
            '': b'## Interface: 11300\n',
        },
        'flavours': {Flavour.VanillaClassic},
    },
}


@pytest.fixture(params=ZIPS.values(), ids=list(ZIPS))
def package_json_less_addon(
    request: pytest.FixtureRequest,
):
    addon = BytesIO()

    with ZipFile(addon, 'w') as file:
        for flavour_suffix, content in request.param['toc_files'].items():
            file.writestr(f'{zip_addon_name}/{zip_addon_name}{flavour_suffix}.toc', content)

    return (addon.getvalue(), request.param['flavours'])


@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [
        {
            f'//api.github.com/repos/{zip_defn.alias}',
            f'//api.github.com/repos/{zip_defn.alias}/releases?per_page=10',
        }
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    'iw_profile_config_values',
    Flavour,
    indirect=True,
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_extracting_flavour_from_zip_contents(
    iw_aresponses: ResponsesMockServer,
    iw_config_ctx: ConfigBoundCtx,
    github_resolver: GithubResolver,
    package_json_less_addon: tuple[bytes, set[Flavour]],
):
    async def handle_request(request: aiohttp.web.Request):
        if aiohttp.hdrs.RANGE in request.headers:
            raise aiohttp.web.HTTPRequestRangeNotSatisfiable

        response = aiohttp.web.Response(body=addon)
        await response.prepare(request)
        return response

    iw_aresponses.add(
        **Route(
            '//api.github.com',
            path_pattern=re.compile(r'^/repos(/[^/]*){2}/releases/assets/'),
            response=handle_request,
        ).to_aresponses_add_args()
    )

    addon, flavours = package_json_less_addon
    try:
        await github_resolver.resolve_one(zip_defn, None)
    except PkgFilesNotMatching:
        assert iw_config_ctx.config.game_flavour not in flavours
    else:
        assert iw_config_ctx.config.game_flavour in flavours


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_repo_with_release_json_release(
    github_resolver: GithubResolver,
):
    defn = Defn('github', 'nebularg/PackagerTest')

    result = await github_resolver.resolve_one(defn, None)
    assert type(result) is Pkg


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_repo_without_releases(
    github_resolver: GithubResolver,
):
    defn = Defn('github', 'AdiAddons/AdiBags')

    with pytest.raises(PkgFilesMissing) as exc_info:
        await github_resolver.resolve_one(defn, None)

    assert exc_info.value.message == 'no releases found'


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_nonexistent_repo(
    github_resolver: GithubResolver,
):
    defn = Defn('github', 'layday/foobar')

    with pytest.raises(PkgNonexistent):
        await github_resolver.resolve_one(defn, None)


@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [
        {
            f'//api.github.com/repos/{packager_test_defn.alias}',
            f'//api.github.com/repos/{packager_test_defn.alias}/releases?per_page=10',
        }
    ],
    indirect=True,
)
@pytest.mark.parametrize('any_flavour', [True, None])
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_any_flavour_strategy(
    iw_aresponses: ResponsesMockServer,
    iw_config_ctx: ConfigBoundCtx,
    github_resolver: GithubResolver,
    any_flavour: Literal[True, None],
):
    opposite_flavour = next(f for f in Flavour if f is not iw_config_ctx.config.game_flavour)
    opposite_interface = next(
        n for r in opposite_flavour.to_flavour_keyed_enum(FlavourVersionRange).value for n in r
    )

    iw_aresponses.add(
        **Route(
            '//api.github.com',
            path_pattern=re.compile(r'^/repos/nebularg/PackagerTest/releases/assets/'),
            response={
                'releases': [
                    {
                        'filename': 'TestGit-v1.9.7.zip',
                        'nolib': False,
                        'metadata': [
                            {
                                'flavor': opposite_flavour,
                                'interface': opposite_interface,
                            }
                        ],
                    }
                ]
            },
        ).to_aresponses_add_args()
    )
    defn = Defn(
        'github',
        'nebularg/PackagerTest',
        strategies=Strategies({Strategy.AnyFlavour: any_flavour}),
    )

    results = await github_resolver.resolve([defn])
    assert type(results[defn]) is (Pkg if any_flavour else PkgFilesNotMatching)


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_changelog_is_data_url(
    github_resolver: GithubResolver,
):
    defn = Defn('github', 'p3lim-wow/Molinari')

    result = await github_resolver.resolve_one(defn, None)
    assert result.changelog_url.startswith('data:,')


@pytest.mark.parametrize(
    ('iw_profile_config_values', 'flavor', 'interface'),
    [
        (Flavour.Retail, 'mainline', 30400),
        (Flavour.Classic, 'cata', 90207),
        (Flavour.VanillaClassic, 'classic', 90207),
    ],
    indirect=('iw_profile_config_values',),
)
@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [
        {
            f'//api.github.com/repos/{packager_test_defn.alias}',
            f'//api.github.com/repos/{packager_test_defn.alias}/releases?per_page=10',
        }
    ],
    indirect=True,
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_mismatched_release_is_skipped_and_logged(
    caplog: pytest.LogCaptureFixture,
    iw_aresponses: ResponsesMockServer,
    iw_config_ctx: ConfigBoundCtx,
    github_resolver: GithubResolver,
    flavor: str,
    interface: int,
):
    iw_aresponses.add(
        **Route(
            '//api.github.com',
            path_pattern=re.compile(r'^/repos/nebularg/PackagerTest/releases/assets/'),
            response={
                'releases': [
                    {
                        'filename': 'TestGit-v1.9.7.zip',
                        'nolib': False,
                        'metadata': [{'flavor': flavor, 'interface': interface}],
                    }
                ]
            },
        ).to_aresponses_add_args()
    )

    with pytest.raises(PkgFilesNotMatching):
        await github_resolver.resolve_one(packager_test_defn, None)

    assert (
        'instawow._sources.github',
        logging.INFO,
        f'flavor and interface mismatch: {interface} not found in '
        f'{[iw_config_ctx.config.game_flavour.to_flavour_keyed_enum(FlavourVersionRange).value]}',
    ) in caplog.record_tuples


@pytest.mark.parametrize(
    ('url', 'extracted_alias'),
    [
        (
            'https://github.com/AdiAddons/AdiButtonAuras',
            'AdiAddons/AdiButtonAuras',
        ),
        (
            'https://github.com/AdiAddons/AdiButtonAuras/releases',
            'AdiAddons/AdiButtonAuras',
        ),
    ],
)
def test_can_extract_alias_from_url(
    github_resolver: GithubResolver,
    url: str,
    extracted_alias: str,
):
    assert github_resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

from collections.abc import Awaitable, Collection, Mapping, Set
from typing import Any, Final, Protocol, TypeAlias, TypeVar

from . import pkg_models
from ._logging import logger
from .definitions import Strategies, Strategy

_T = TypeVar('_T')

AnyResult: TypeAlias = '_T | ManagerError | InternalError'


class Result(Protocol):  # pragma: no cover
    @property
    def message(self) -> str: ...


class _SuccessResult:
    status: Final = 'success'


class PkgInstalled(Result, _SuccessResult):
    def __init__(self, pkg: pkg_models.Pkg, *, dry_run: bool = False) -> None:
        super().__init__()
        self.pkg = pkg
        self.dry_run = dry_run

    @property
    def message(self) -> str:
        return f'{"would have installed" if self.dry_run else "installed"} {self.pkg.version}'


class PkgUpdated(Result, _SuccessResult):
    def __init__(
        self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg, *, dry_run: bool = False
    ) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg
        self.dry_run = dry_run

    @property
    def message(self) -> str:
        message = (
            f'{"would have updated" if self.dry_run else "updated"}'
            f' {self.old_pkg.version} to {self.new_pkg.version}'
        )

        if self.old_pkg.slug != self.new_pkg.slug:
            message += f' with new slug {self.new_pkg.slug!r}'

        if self.old_pkg.options != self.new_pkg.options:
            old_strategies = self.old_pkg.to_defn().strategies
            new_strategies = self.new_pkg.to_defn().strategies
            message += ' with new strategies: '
            message += '; '.join(
                f'{s}={v!r}' for s, v in new_strategies.items() - old_strategies.items()
            )

        return message


class PkgRemoved(Result, _SuccessResult):
    def __init__(self, old_pkg: pkg_models.Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg

    @property
    def message(self) -> str:
        return 'removed'


class ManagerError(Result, Exception):
    status: Final = 'failure'

    @property
    def message(self) -> str:
        raise NotImplementedError


class PkgAlreadyInstalled(ManagerError):
    @property
    def message(self) -> str:
        return 'package already installed'


class PkgConflictsWithInstalled(ManagerError):
    def __init__(self, conflicting_pkgs: Collection[Mapping[str, Any]]) -> None:
        super().__init__()
        self.conflicting_pkgs = conflicting_pkgs

    @property
    def message(self) -> str:
        return (
            'package folders conflict with installed package'
            + ('s ' if len(self.conflicting_pkgs) > 1 else ' ')
            + ', '.join(f'{c["name"]} ({c["source"]}:{c["id"]})' for c in self.conflicting_pkgs)
        )


class PkgConflictsWithUnreconciled(ManagerError):
    def __init__(self, folders: Set[str]) -> None:
        super().__init__()
        self.folders = folders

    @property
    def message(self) -> str:
        folders = ', '.join(f"'{f}'" for f in self.folders)
        return f'package folders conflict with {folders}'


class PkgNonexistent(ManagerError):
    @property
    def message(self) -> str:
        return 'package does not exist'


class PkgFilesMissing(ManagerError):
    def __init__(self, custom_message: str | None = None) -> None:
        super().__init__()
        self._custom_message = custom_message

    @property
    def message(self) -> str:
        return self._custom_message or 'no files are available for download'


class PkgFilesNotMatching(ManagerError):
    def __init__(self, strategies: Strategies) -> None:
        super().__init__()
        self.strategies = strategies

    @property
    def message(self) -> str:
        return 'no files found for: ' + '; '.join(f'{s}={v!r}' for s, v in self.strategies.items())


class PkgNotInstalled(ManagerError):
    @property
    def message(self) -> str:
        return 'package is not installed'


class PkgSourceInvalid(ManagerError):
    @property
    def message(self) -> str:
        return 'package source is invalid'


class PkgUpToDate(ManagerError):
    def __init__(self, is_pinned: bool) -> None:
        super().__init__()
        self.is_pinned = is_pinned

    @property
    def message(self) -> str:
        return f'package is {"pinned" if self.is_pinned else "up to date"}'


class PkgStrategiesUnsupported(ManagerError):
    def __init__(self, strategies: Collection[Strategy]) -> None:
        super().__init__()
        self.strategies = strategies

    @property
    def message(self) -> str:
        return f'strategies are not valid for source: {", ".join(sorted(self.strategies))}'


class InternalError(Result, Exception):
    status: Final = 'error'

    @property
    def message(self) -> str:
        return f'internal error: "{self.args[0]}"'


async def resultify_async_exc(
    awaitable: Awaitable[_T],
) -> AnyResult[_T]:
    "Capture and log an exception raised in a coroutine."
    try:
        return await awaitable
    except (ManagerError, InternalError) as error:
        return error
    except BaseException as error:
        logger.exception('unclassed error')
        return InternalError(error)



# pyright: reportUnknownArgumentType = false
# pyright: reportUnknownMemberType   = false

from __future__ import annotations

import asyncio
import concurrent.futures
import json
import sys
import threading
from collections.abc import Callable
from functools import partial

import toga
import toga.constants
import toga.style.pack

from instawow._utils.compat import StrEnum

from . import json_rpc_server

_loop_factory = asyncio.DefaultEventLoopPolicy().new_event_loop


class _TogaSimulateKeypressAction(StrEnum):
    ToggleSearchFilter = 'toggleSearchFilter'
    ActivateViewInstalled = 'activateViewInstalled'
    ActivateViewReconcile = 'activateViewReconcile'
    ActivateViewSearch = 'activateViewSearch'


class _App(toga.App):
    def __start_json_rpc_server(self, on_started: Callable[[str], None]):
        def start_json_rpc_server():
            async def main():
                async with json_rpc_server.run_web_app(
                    await json_rpc_server.create_web_app(self)
                ) as server_url:
                    server_url_future.set_result(str(server_url))
                    await wait_future

            loop.run_until_complete(main())

        def stop_json_rpc_server():
            loop.call_soon_threadsafe(wait_future.set_result, None)
            json_rpc_server_thread.join()

        loop = _loop_factory()
        wait_future = loop.create_future()

        server_url_future = concurrent.futures.Future[str]()
        server_url_future.add_done_callback(lambda f: on_started(f.result()))

        json_rpc_server_thread = threading.Thread(
            target=start_json_rpc_server, name='_json_rpc_server'
        )
        json_rpc_server_thread.start()

        def on_app_exit(app: toga.App, **kwargs: object):
            stop_json_rpc_server()
            return True

        self.on_exit = on_app_exit

    def startup(self) -> None:
        self.main_window = toga.MainWindow(title=self.formal_name, size=(800, 600))

        if sys.platform == 'win32':
            import ctypes

            # Enable high DPI support.
            ctypes.windll.user32.SetProcessDPIAware()

        def dispatch_js_keyboard_event(command: toga.Command, **kwargs: object):
            event_args = json.dumps(
                {'detail': {'action': kwargs['action']}, 'bubbles': True, 'cancelable': True}
            )
            web_view.evaluate_javascript(
                f'document.dispatchEvent(new CustomEvent("togaSimulateKeypress", {event_args}));'
            )

            return True

        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ToggleSearchFilter,
                ),
                text='Toggle Search Filter',
                shortcut=toga.Key.MOD_1 + toga.Key.G,
                group=toga.Group.EDIT,  # pyright: ignore[reportArgumentType]
                section=20,
                order=10,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewInstalled,
                ),
                text='Installed',
                shortcut=toga.Key.MOD_1 + toga.Key.L,
                group=toga.Group.WINDOW,  # pyright: ignore[reportArgumentType]
                section=20,
                order=10,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewReconcile,
                ),
                text='Unreconciled',
                group=toga.Group.WINDOW,  # pyright: ignore[reportArgumentType]
                section=20,
                order=20,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewSearch,
                ),
                text='Search',
                shortcut=toga.Key.MOD_1 + toga.Key.F,
                group=toga.Group.WINDOW,  # pyright: ignore[reportArgumentType]
                section=20,
                order=30,
            ),
        )

        web_view = toga.WebView(style=toga.style.pack.Pack(flex=1))

        self.__start_json_rpc_server(partial(setattr, web_view, 'url'))

        self.main_window.content = web_view
        self.main_window.show()


def make_app(version: str) -> toga.App:
    return _App(
        formal_name='instawow-gui',
        app_id='org.instawow.instawow_gui',
        app_name='instawow_gui',
        icon='resources/instawow_gui',
        version=version,
    )



from __future__ import annotations

from collections.abc import Hashable, Iterable, Iterator, Mapping
from functools import partial
from typing import Literal, overload

import attrs
from typing_extensions import Self
from yarl import URL

from ._utils.compat import StrEnum, fauxfrozen


class Strategy(StrEnum):
    AnyFlavour = 'any_flavour'
    AnyReleaseType = 'any_release_type'
    VersionEq = 'version_eq'


class ChangelogFormat(StrEnum):
    Html = 'html'
    Markdown = 'markdown'
    Raw = 'raw'


@fauxfrozen
class SourceMetadata:
    id: str
    name: str
    strategies: frozenset[Strategy]
    changelog_format: ChangelogFormat
    addon_toc_key: str | None


class Strategies(Mapping[Strategy, object], Hashable):
    initialised = True

    def __init__(self, entries: Mapping[Strategy, object] = {}) -> None:
        self._entries = dict.fromkeys(Strategy) | dict(entries)

    @overload
    def __getitem__(self, key: Literal[Strategy.AnyFlavour], /) -> Literal[True] | None: ...
    @overload
    def __getitem__(self, key: Literal[Strategy.AnyReleaseType], /) -> Literal[True] | None: ...
    @overload
    def __getitem__(self, key: Literal[Strategy.VersionEq], /) -> str | None: ...
    @overload
    def __getitem__(self, key: Strategy, /) -> object: ...
    def __getitem__(self, key: Strategy, /) -> object:
        return self._entries[key]

    def __hash__(self) -> int:
        return hash(frozenset(self._entries.items()))

    def __iter__(self) -> Iterator[Strategy]:
        return iter(self._entries)

    def __len__(self) -> int:
        return len(self._entries)

    @property
    def filled(self) -> dict[Strategy, object]:
        return {s: v for s, v in self._entries.items() if v is not None}


class _UninitialisedStrategies(Strategies):
    initialised = False


_STRATEGY_SEP = ','
_STRATEGY_VALUE_SEP = '='


@fauxfrozen
class Defn:
    source: str
    alias: str
    id: str | None = None
    strategies: Strategies = _UninitialisedStrategies()

    @classmethod
    def from_uri(
        cls,
        uri: str,
        *,
        known_sources: Iterable[str],
        allow_unsourced: bool,
    ) -> Self:
        """Construct a ``Defn`` from a URI."""
        url = URL(uri)

        if url.scheme not in known_sources:
            if not allow_unsourced:
                raise ValueError(f'Unable to extract source from {uri}')

            make_cls = partial(cls, source='', alias=uri)
        else:
            make_cls = partial(cls, source=url.scheme, alias=url.path)

        if url.fragment == _STRATEGY_VALUE_SEP:
            make_cls = partial(make_cls, strategies=Strategies())
        elif url.fragment:
            strategy_values = {
                s: v
                for f in url.fragment.split(_STRATEGY_SEP)
                for s, _, v in (f.partition(_STRATEGY_VALUE_SEP),)
                if s
            }

            unknown_strategies = strategy_values.keys() - set(Strategy)
            if unknown_strategies:
                raise ValueError(f'Unknown strategies: {", ".join(unknown_strategies)}')

            make_cls = partial(
                make_cls,
                strategies=Strategies(
                    {
                        Strategy.AnyFlavour: Strategy.AnyFlavour in strategy_values or None,
                        Strategy.AnyReleaseType: Strategy.AnyReleaseType in strategy_values
                        or None,
                        Strategy.VersionEq: strategy_values.get(Strategy.VersionEq),
                    }
                ),
            )

        return make_cls()

    def as_uri(self, *, alias_is_id: bool = False, include_strategies: bool = False) -> str:
        uri = f'{self.source}:{self.id if alias_is_id else self.alias}'

        if include_strategies:
            filled = self.strategies.filled
            if filled:
                uri += f"""#{_STRATEGY_SEP.join(
                    s if v is True else f"{s}={v}" for s, v in filled.items()
                )}"""

        return uri

    def with_default_strategy_set(self) -> Self:
        return attrs.evolve(
            self,
            strategies=Strategies(),
        )

    def with_version(self, version: str) -> Self:
        return attrs.evolve(
            self,
            strategies=Strategies({**self.strategies, Strategy.VersionEq: version}),
        )
